import Fudgets
import ExtraFudgets

main = fudlogue $ shellF "Roll" rollF

rollF = revVBoxF (scoreF >==< boardF)

data Scores = Win | Loose deriving (Eq)

scoreF =
    spacerF centerS $
    hBoxF $
    listF [(Win,scoreDispF "You win"),
           (Loose,scoreDispF "You loose")]
  where
    scoreDispF lbl = lbl `labLeftOfF` intDispF >==< countF
       where
	countF = mapstateF inc 0
        inc n _ = (n+1,[n+1])

boardF = eventF eventmask (graphicsBgF "yellow" boardD ballF) >==< tickF
  where
    eventmask = [PointerMotionMask]
    tickF = startupF [Just (dt,1000)] timerF
 
dt = argReadKey "dt" orig_dt
orig_dt = 25 -- time step in the original program

boardD = stackD [dangersD,bouncesD,startAndFinishD,linesD]
  where
    fd = atomicD . FixD size
    dangersD = fgD "black" $ fd $ map FillRectangle dangers
    bouncesD = fgD "blue"  $ fd $ map FillRectangle bounces
    startAndFinishD = fd [DrawString start "Start",DrawString finish "Finish"]
    linesD = fd $ map DrawLine lines'


-- Geometrical info:
magicx = 10
magicy = 10
width = 600+magicx
height = 600+magicy
size = Point width height -- board size
dist = 300
ballsize = 40
halfbsize = ballsize `div` 2
border = 2

dangers = [rR 0 0 100 100, rR 500 250 100 100, rR 300 500 100 100]
bounces = [rR 250 100 300 50, rR 100 100 50 200, rR 250 300 100 100,
	   rR 0 450 200 50]
lines' = [lL dist border (width-dist) border,
          lL dist (height-border) (width-dist) (height-border)]
start = Point 550 25
finish = Point 10 580


ballF = shapedObjectF "red" [fillCircle origin ballsize] startrect ballSP
  where
    ballSP = run state0

    run s = getSP $ either handleEvent handleTick
      where
        same = run s
	handleTick _ = move

	handleEvent event =
	  case event of
	    MotionNotify { pos=pos } -> tilt pos
	    _ -> same

        tilt (Point tx ty) =
          run (s { tiltx = real (tx-middlex), tilty = real (ty-middley) })

        move_ball s' =
	    if p' /= ballpos s
	    then putSP (Left p') $
	         run s'
	    else run s'
          where p' = ballpos s'

        move =
	    if any (fall_into (ix s1+halfbsize,iy s1+halfbsize)) dangers
	    then you Loose
	    else if ix s1 < xwin && iy s1 > ywin
	         then you Win
	         else move_ball s'
 	  where
             you outcome =
	       putSP (Right (outcome,Tick)) $
	       move_ball state0

	     s1 = s { dx = newdelta (tiltx s) (dx s),
		      dy = newdelta (tilty s) (dy s),
		      rx = rx s + dx s1 * (dt / orig_dt),
		      ry = ry s + dy s1 * (dt / orig_dt) }
	     s2 = if rx s1 > rx s
	          then if rx s1 > xmax
		       then s1 { rx=xmax, dx = -dx s1 }
	               else foldr (bounce_left (rx s) (ry s)) s1 bounces
	          else if rx s1 < rx s
                       then if rx s1 < 0
	                    then s1 { rx=0, dx = -dx s1 }
			    else foldr (bounce_right (rx s) (ry s)) s1 bounces
		       else s1
	     s' = if ry s2 > ry s
	          then if ry s2 > ymax
		       then s2 { ry=ymax, dy = -dy s2 }
	               else foldr (bounce_above (rx s) (ry s)) s2 bounces
	          else if ry s2 < ry s
                       then if ry s2 < 0
	                    then s2 { ry=0, dy = -dy s2 }
			    else foldr (bounce_below (rx s) (ry s)) s2 bounces
		       else s2

data BallState = B { rx,ry,tiltx,tilty,dx,dy::Double }
ix = round . rx
iy = round . ry
state0 = B xstart ystart 0 0 0.0 0.0
ballpos s = Point (ix s) (iy s)

xmax = real (width - ballsize)
ymax = real (height - ballsize)
xstart = xmax - ballsize - 10  -- magicx?
ystart = ballsize + 10  -- magicy?
startpos = Point xstart ystart
startrect = Rect startpos (ballsize+2) -- +2?
middlex = width `div` 2
middley = height `div` 2
xwin = halfbsize
ywin = round ymax - halfbsize

newdelta ftilt delta =
    if ftilt > 0.0 && delta < 0.0 then
      delta + ftilt / 500.0
    else if ftilt < 0.0 && delta > 0.0 then
      delta + ftilt / 500.0
    else if delta > height then
      height
    else if delta < - height then
       - height
    else
      delta + ftilt / 2000.0

real i = fromIntegral i

-- fall_into (xb,yb) r = Point xb yb `inRect` r

fall_into (xb,yb) (Rect (Point x y) (Point w h)) =
    x < xb && xb < x1 && y < yb && yb < y1
  where
    x1 = x+w
    y1 = y+h

bounce_left xb yb r s@(B{rx=xn, ry=yn}) =
    if xb <= a && a < xn && b <= bn && bn <= d
    then s { rx = a, {-ry = bn,-} dx = -dx s }
    else s
  where
    bn = ((a - xb) * (yn - yb)) / (xn - xb) + yb -- using laziness !
    (a,b,_,d) = bouncepos r

bounce_right xb yb r s@(B{rx=xn, ry=yn}) =
    if xb >= c && c > xn && b <= bn && bn <= d
    then s { rx = c, {-ry = bn,-} dx = -dx s }
    else s
  where
    bn = ((c - xb) * (yn - yb)) / (xn - xb) + yb -- using laziness !
    (_,b,c,d) = bouncepos r

bounce_above xb yb r s@(B{rx=xn, ry=yn}) =
    if yb <= b && b < yn && a <= an && an <= c
    then s { {-rx = an,-} ry = b, dy = -dy s }
    else s
  where
    an = ((b - yb) * (xn - xb)) / (yn - yb) + xb -- using laziness !
    (a,b,c,_) = bouncepos r

bounce_below xb yb r s@(B{rx=xn, ry=yn}) =
    if yb >= d && d > yn && a <= an && an <= c
    then s { {-rx = an,-} ry = d, dy = -dy s }
    else s
  where
    an = ((d - yb) * (xn - xb)) / (yn - yb) + xb -- using laziness !
    (a,_,c,d) = bouncepos r

bouncepos (Rect (Point x y) (Point w h)) = (a,b,c,d)
  where
    a = real (x - ballsize)
    b = real (y - ballsize)
    c = real (x + w)
    d = real (y + h)
