module
#include "q.t"
#include "utils.t"
#include "move.t"

export	wormno,wormhead,wormbody,wormdir,wormgrow,	-- selectors
	wormscore,
	updhead,updbody,upddir,updgrow,updscore,	-- update functions
	newworm,					-- constructor
	insideworm,					-- predicates
	cutwormtail,movewormhead,			-- special modifiers
	changedir,storedirchange
	;

-- This is an abstract data type for worms. It may seem meaningless to use
-- an abstract type for such a simple structure, but I think it is motivated,
-- since LML lacks possibilities to update single fields of a data structure
-- without having to explicitly name and copy ALL fields.

-- Disadvantage: pattern matching can't be used to get the parts of a worm
-- Advantage: you can extend the type with more fields without having to
--	      make changes to the code using the type (except where the new
--	      fields are to be used)

local
  type WORM = Worm Int			-- worm number
		   (Int#Int)		-- position of head
		   (QUEUE (Int#Int))	-- worm body (including head)
		   (Int#Int)		-- current direction
		   Int			-- how much to grow
		   Int			-- score
		   (QUEUE ((Int#Int)->(Int#Int)))
					-- outstanding direction changes

in
  rec wormno    (Worm no _ _ _ _ _ _)   =no
  and wormhead  (Worm _ head _ _ _ _ _) =head
  and wormbody  (Worm _ _ body _ _ _ _) =body
  and wormdir   (Worm _ _ _ dir _ _ _)  =dir
  and wormgrow  (Worm _ _ _ _ grow _ _) =grow
  and wormscore (Worm _ _ _ _ _ score _)=score
  and wormdirq  (Worm _ _ _ _ _ _ dirq) =dirq
  and updhead  f (Worm n h b d g s q) =Worm n (f h) b d g s q
  and updbody  f (Worm n h b d g s q) =Worm n h (f b) d g s q
  and upddir   f (Worm n h b d g s q) =Worm n h b (f d) g s q
  and updgrow  f (Worm n h b d g s q) =Worm n h b d (f g) s q
  and updscore f (Worm n h b d g s q) =Worm n h b d g (f s) q
  and upddriq  f (Worm n h b d g s q) =Worm n h b d g s (f q)

  and newworm no head dir grow score=
    Worm no head (Enter Empty head) dir grow score Empty

  and insideworm pos worm= qmember (wormbody worm) pos

  and cutwormtail (Worm n h body d g s q)=
    let (tail,rest)=qremove body
    in (tail,Worm n h rest d g s q)

  and movewormhead (Worm n _ body d g s q) newhead=
    let newbody = (Enter body newhead)
    in mwh (Worm n newhead) d g s q newbody 
  and mwh w d g s q n=w n d g s q
/*
  and movewormhead newhead (Worm n head body d g s q)=
    Worm n newhead (Enter body newhead) d g s q
*/

  and changedir (worm as Worm n h b dir g s dirchangeq)=
    if isempty dirchangeq
    then worm
    else let (dirchange,rest)=qremove dirchangeq
	 in let newdir=dirchange dir
         in (Worm n h b (if newdir=negv dir
			 then dir
			 else newdir) g s rest)

  and storedirchange dirchange (Worm n h b d g s dirchangeq)=
    Worm n h b d g s (Enter dirchangeq dirchange)

end -- local

end -- module
