#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sgtty.h>

#include "encoding.h"

static void error(char *s)
{
  perror(s);
  exit(1);
}

static int opensocket(char *host,int port)
{
  int s;
  struct hostent *he;
  static struct sockaddr_in sa = { AF_INET };

  printf("gehostbyname...\n");
  he=gethostbyname(host);
  if(!he) error(host);
  printf("IP address %8x\n",*(long *)(he->h_addr));
  printf("Creating socket...\n");
  s=socket(he->h_addrtype,SOCK_STREAM,0);
  if(s<0) error("socket");
  bcopy(he->h_addr,(char *)&sa.sin_addr,he->h_length);
  sa.sin_port=htons(port);
  printf("Connecting...\n");
  if(connect(s,&sa,sizeof sa)) error("connect");
  return s;
}

static void changemode(int s)
{
  struct sgttyb buf;

  if(gtty(s,&buf)) error("gtty");
  printf("sg_flags=%8x\n",buf.sg_flags);
  buf.sg_flags= (buf.sg_flags | CBREAK) & ~ECHO;
  if(stty(s,&buf)) error("stty");
}

static int get(int s)
{
  char c;
  read(s,&c,1);
  return c;
}

#define CSI "\033["	/* control sequence introducer (ANSI terminals) */

static void decode(int s)
{
   char c;
   int x,y;

   c=get(s);
   switch(c) {
     case STRTAG:
       while((c=get(s))!=STREND) putchar(c);
       break;
     case FLUSH:
       fflush(stdout);
       break;
     case MOVETAG:
       x=get(s); y=get(s);
       printf("%s%d;%dH",CSI,y,x);
       /*
       printf(CSI "%d;%dH",y+1,x+1);
       */
       break;
     case CLEAR:
       printf("%sH%sJ",CSI,CSI);
       break;
   }
}

static void play(int s,int t)
{
  fd_set rd;
  int n;

  FD_ZERO(&rd);
  while(1) {
    FD_SET(s,&rd);
    FD_SET(t,&rd);
    n=select(getdtablesize(),&rd,NULL,NULL,NULL);
    if(n<0) error("select");
    if(FD_ISSET(t,&rd)) {
      char c;
      c=get(t);
      write(s,&c,1);
    }
    if(FD_ISSET(s,&rd)) decode(s);
  }
}

void main(int argc,char **argv)
{
  int s,t;

  s=opensocket(argc>1 ? argv[1] : "lips",10309);
  t=fileno(stdin);
  changemode(t);
  printf("OK!\n");
  fflush(stdout);
  sleep(2);
  write(s,"C program\n",10);
  play(s,t);
}
