import Text.PrettyPrint
import Data.List(nub,partition,groupBy,sortBy,(\\))
import Data.Char(isSpace,isUpper,toLower)
--import IOUtil(getEnvi)
import Data.Maybe(fromMaybe)
import Unsafe(unsafeOpenFile)
import Syntax
import ParseFile
import PP
import SyntaxUtils
import HtmlUtils
import Utils(mix,mapPartition,pairwith)
import Data.Bifunctor(first)
--import ListUtil(mapFst) -- assoc
--import HO(apFst)
import System.Time(getClockTime,toCalendarTime,calendarTimeToString)
import Debug.Trace(trace)

import qualified Data.Map as Map
type Map a b = Map.Map a b

--fudgetDir = "/usr/src/local/lml/lmlx/"
fudgetDir = "../../"
fudgetDocDir = fudgetDir++"Doc/latexinfo/"
--fudgetLibDir = fudgetDir++"lib/"

-- "Intertface" files created by mksections
fudgetsHi = "AllFudgets.hi"
contribHi = "ContribFudgets.hi"
smallfudgetsHi = "Fudgets.hi"

fudgetManDir = "FudgetMan/" -- directory for HTML output
fudgetManIndex = fudgetManDir++fullIndexFile
fudgetManBeginnerIndex = fudgetManDir++"beginner.html"
fudgetManSmallIndex = fudgetManDir++"small.html"
fudgetManOldIndex = fudgetManDir++"old.html"

libraryName = fromMaybe "Fudget Library" $ Nothing -- getEnvi "LIBNAME"
fullIndexFile = fromMaybe "index.html" $ Nothing -- getEnvi "FULLINDEXFILE"

main =
  do docentries <- (doc . lines) `fmap` getContents
     --putStr (show docentries) -- debugging
     writeIt docentries

writeIt docentries0 =
  do let names = map fst docentries
         crossrefs = refpage (map classfree names)
	 docentries = map (first (escsynopsis crossrefs)) docentries0
     today <- getTodaysDate
     writeIndices today ("",crossrefs) names
     let (images,docentries'') = unzip $ map fexamples docentries
     writeFile "fexample.hs2" (fexample2hs (concat images)) 
     writeEntries crossrefs docentries''
  where
    escsynopsis es (this,others,cs,ws) = (this,others,cs,synopsis)
      where synopsis = unwords ws
                       -- escapechars ("",es) (unwords ws)

    getTodaysDate =
       calendarTimeToString `fmap` (getClockTime >>= toCalendarTime)

writeEntries crossrefs = mapM_ (writeEntry crossrefs)

writeEntry crossrefs (n@(this,_,_,_),ls) =
  do let file = fudgetManDir++this++".html"
     writeDoc file (docentry crossrefs n ls)

doc (('$':'D':'o':'c':' ':cs):ls)=
       let this:others = words cs
	   (ce,ls')    = enddoc [] ls
	   (c,e0)      = docclass ce
	   (sy,e)      = docsynopsis e0
       in ((this,others,c,sy),e):doc ls'
doc (_:ls) = doc ls
doc [] = []

enddoc es (('$':'E':'n':'d':'d':'o':'c':_):ls) = (reverse es,ls)
enddoc es (l@('$':'D':'o':'c':_):ls) = trace "*** missing $Enddoc" (reverse es,l:ls)
enddoc es (('\\':'c':[]   ):ls) = enddoc es ls
enddoc es (('\\':'c':t:_):ls) | isSpace t = enddoc es ls
enddoc es (l:ls) = enddoc (l:es) ls
enddoc es [] = (reverse ("$Enddoc":es),[])

docclass = extractlines "$C"
docsynopsis = extractlines "$Synopsis"

extractlines w1 = first concat . mapPartition match
  where match l =
           case words l of
	     w:ws | w==w1 -> Just ws
	     _ -> Nothing

classfree (this,others,c,syn) = (this,others)

docentry es (this,others,_,synopsis) ls =
    on "!DOCTYPE html" $$
    on "HTML" $$
    docfmt crossrefs (this:others) synopsis $$
    usage this argnames $$
    vmap (para args crossrefs) (arghdr ls) $$
    vcat [on "!--#include file=\"footer.html\"--",off "BODY",off "HTML"]
  where
    crossrefs = (this,es)
    args = zip argnames argtypes
    argnames = [a|"$Arg":a:_<-map words ls]
    argtypes = case Map.lookup this values of
                 Just (Sig _ (CT _ t)) -> funArgs t
		 _ -> []

para args es "" = on "P"
para args es ('$':cs) | kind/="this" = parafmt args es kind txt
    where (kind,txt) = takeWord cs
para args es cs = escapechars es cs

parafmt args es kind txt =
    maybe defaultpara (\f->f (args,es) txt) $ lookup kind parafmts
--  assoc (\f->f (args,es) txt) defaultpara parafmts kind
  where
    defaultpara = wrap "H2" (plain kind) $$ escapechars es txt

    parafmts = [("Input",iofmt "Input"),
		("Output",iofmt "Output"),
		("Type",\es->typefmt es.fst.firstWord),
		("Arguments",const arghdrfmt),
		("Argend",const argendfmt),
                ("Arg", argfmt),
		("Example", examplefmt),
		("MExample",examplefmt),
		("FExample",examplefmt),
		("See",seefmt)
	       ]

docfmt es ns synopsis =
    title2 (libraryName ++ " Manual: "++tit) ("¤ "++tit) $ 
    plain synopsis $$
    wrap "H2" (text "Types") $$
    on "PRE"<>vmap (idDef es) ns<>off "PRE"
  where
    tit = mix (take maxnames ns++etal) ", "
    etal = if null (drop maxnames ns) then [] else ["et al"]
    maxnames = 2

idDef es name@(c:_) =
    if isUpper c
    then ppTypeDef es (typeDef name) $$ vmap (ppTopDecl es) instances
    else ppValueDef es (valueDef name) $$ infixes
  where
    instances = nub [d | Interface _ ds<-fudgetsHiLines,
                         d@(Instance _ (Con cl (ConT (Con n _):_)))<-ds,
			 n==name || cl==name,cl/="Eval"]
    infixes = case lookup name infixdecs of
                  Just d -> ppTopDecl es d
		  _ -> empty

title txt = title2 txt txt

title2 title header rest =
    on "HEAD" $$
    on "META charset=\"UTF-8\"" $$
    stylelink $$
    wrap "TITLE" (plain title) $$
    off "HEAD" $$
    on "BODY" $$
    on "!--#include file=\"header.html\"--" $$
    wrap "H1" (plain header) $$
    rest
  where
    stylelink = tag "LINK rel=stylesheet href=\"manual.css\""

usage _ [] = empty
usage this args =
  wrap "H2" (text "Synopsis") $$ wrap "PRE" (hsep (map plain (this:args)))

arghdr [] = []
arghdr (l:ls) =
  case words l of
    "$Arg":_ -> "$Arguments":l:argend ls
    _ -> l:arghdr ls

argend [] = ["$Argend"]
argend lls@(l:ls) =
  case words l of
    ('$':w):_ | w `notElem` ["this","Arg"] ->
                  if w=="Argend"
		  then lls
		  else "$Argend":lls
    _ -> l:argend ls

typefmt (_,es) t = wrap "PRE" (ppTypeDef es (typeDef t))
iofmt h (_,es) txt= wrap "H3" (plain h) $$ escapechars es txt

argfmt (args,crossrefs) txt =
  let (arg,ws) = firstWord txt
    --typ = assoc (ppType crossrefs) (text "??") args arg
      typ = maybe (text "??") (ppType crossrefs) $ lookup arg args
  in (wrap "DT".wrap "CODE") (wrap "B" (text arg) <+> text"::"<+>typ) $$
     nest 4 (on "DD" <> escapechars crossrefs ws)

arghdrfmt _ = wrap "H2" (text "Arguments") $$ on "DL"
argendfmt _ = off "DL"

examplefmt (_,es) txt = wrap "H2" (text "Example") $$
                        wrap "PRE" (escapechars es txt)
seefmt (_,es) txt = wrap "H2" (text "See Also") $$ escapechars es txt
--enddocfmt _ = ["<HR>"]

escapechars es = chars es -- . replacethis es (wrap "CODE" (plain this))

replacethis es@(this,_) cs =
  case cs of
    "" -> empty
    '$':'t':'h':'i':'s':rest -> wrap "CODE" (plain this)<>replacethis es rest
    c:rest -> plain [c]<>replacethis es rest

chars es "" = empty
chars es cs =
  case break (=='\\') cs of
    (_,"") -> replacethis es cs
    (before,'\\':rest) ->
      case break (\c->c=='{' || isSpace c) rest of
        (_,"") -> replacethis es cs
	(cmd,rest@(c:_)) | isSpace c ->
	   replacethis es before<>charcmd cmd<>chars es rest
	(cmd,'{':rest) ->
	  case break (=='}') rest of
	    (_,"") -> plain cs
	    (arg,'}':rest) -> replacethis es before<>
                              charfmt es cmd arg<>chars es rest

charcmd cmd = --assoc on (plain ('\\':cmd)) charcmds cmd
              maybe (plain ('\\':cmd)) on $ lookup cmd charcmds
charcmds = [("item","LI")]

charfmt es cmd arg =
    --assoc (\f->f es arg) (defaultchars es cmd arg) charfmts cmd
    maybe (defaultchars es cmd arg) (\f->f es arg) $ lookup cmd charfmts
  where
    charfmts = [("emph",const $ wrap "EM".replacethis es),
                ("code",const $ wrap "CODE".replacethis es),
                ("strong",const $ strong.replacethis es),
                ("key",const $ wrap "KBD".replacethis es),
                ("kbd",const $ wrap "TT".replacethis es),
                ("xref",ppName),
                ("link",const extref),
                ("input",inputchars),
                ("image",const imgsrc),
                ("begin",const beginchars),
                ("end",const endchars)
               ]
{-
refpoint arg =
  let a = escape arg
  in on ("A NAME="++dq a)<>plain a<>off "A"
-}
intref arg = link ("#"++arg) (plain arg)

extref arg =
  let (url,linktxt) = takeWord arg
      link' = if all isSpace linktxt then url else linktxt
  in link url (plain link')

inputchars es arg =
  case unsafeOpenFile (fudgetDocDir++arg++".tex") of
    Left _ -> plain ("\\input{"++arg++"}")
    Right s -> escapechars es s -- "$this" is not present in s...

beginchars = --assoc on empty charfmtcmds
             maybe empty on . flip lookup charfmtcmds
endchars = --assoc off empty charfmtcmds
           maybe empty off . flip lookup charfmtcmds

charfmtcmds = [("example","PRE"),
               ("itemize","UL"),
               ("enumerate","OL")
	      ]

defaultchars es cmd arg = plain ('\\':cmd)<>braces (replacethis es arg)

takeWord = break isSpace
firstWord = takeWord . dropWhile isSpace

smallNames = [defName d | Interface _ ds <- optReadHiFile smallfudgetsHi, d<-ds]

fudgetsHiLines =
  concatMap splitAllFudgets (readHiFile fudgetsHi) ++ readHiFile contribHi

optReadHiFile = readHiFile' (const [])
readHiFile = readHiFile' error

readHiFile' error f =
   trace ("Reading "++f) $
   case unsafeOpenFile f of
     Left s -> error (f++": "++s)
     Right s -> case parseFile s of
                  Right i -> trace ("OK parse of "++f) i
		  --Left err -> error (f++": syntax error: "++show err)
		  Left (expected,got) ->
                     error (f++": syntax error, expected one of: "++
			    unwords expected++", but got "++take 100 (unwords got))

splitAllFudgets (Interface "AllFudgets" ds) =
  [Interface s [d | d<-ds,defName d `elem` ns] | (s,ns) <- sectionNames]

spiitAllFudgets iface = [iface]

sectionNames = [(s,readLines ("sections/"++s++".names"))|s<-sections]
  where
    sections = readLines "sections/list"

    readLines f = case unsafeOpenFile f of
                    Left e -> error (f++": "++e)
                    Right s -> lines s

types = [(defName d,d) | Interface _ ds <- fudgetsHiLines, d <- ds, isType d]

values =
    Map.fromList
      [(name,s) | Interface _ ds <- fudgetsHiLines,
                  Value s@(Sig name _) <- ds]

infixdecs = [(name,d) | Interface _ ds <- fudgetsHiLines,
                         d@(Infix _ _ name)<-ds]

--typeDef t = assoc id ("type "++t++" = ??") types t

typeDef t =
    (t,
    case [d | (n,d)<-types,n==t] of
      [] -> Nothing
      d@(Data _ cs):ds ->
        case cs of
	  Just _ -> Just d
	  Nothing ->
	    case [d | d@(Data _ (Just _)) <- ds] of
	      [] -> Just d
	      d:_ -> Just d
      d:_ -> Just d)

valueDef v = (v,Map.lookup v values)

writeIndices today crossrefs entries =
  do --appendChan stdout (show beginnerHiLines) -- debugging
     writeDoc fudgetManBeginnerIndex beginnerIndex
     writeDoc fudgetManSmallIndex smallIndex
     writeDoc fudgetManIndex fullIndex
     writeDoc fudgetManOldIndex oldIndex
  where
    fullIndex = ixFile fullHiLines "Full Index"
    beginnerIndex = ixFile beginnerHiLines "Beginner's Index"
    smallIndex = ixFile smallHiLines "Programmer's Index"
    oldIndex = ixFile oldHiLines "Index of obsolete stuff"
    ixFile hi ixn = fudgetIndex today hi ixn crossrefs synopses
    beginnerNames = [name|(this,others,cs,_)<-entries,
                     "beginner" `elem` cs, name<-this:others]
    synopses = Map.fromList
                  [(name,synopsis)|(this,others,_,synopsis)<-entries,
		                   not (null synopsis),
		                   name<-this:others]
    fullHiLines = filterHiLines (const True)
    beginnerHiLines = filterHiLines (`elem` beginnerNames)
    smallHiLines = filterHiLines (`elem` smallNames')
    oldHiLines = filterHiLines (`elem` oldNames)
    smallNames' = smallNames \\ oldNames
    oldNames = [name | (this,others,cs,_)<-entries,
		       "old" `elem` cs,
		       name<-this:others]
    filterHiLines include =
      [Interface n ds | Interface n ds0 <- fudgetsHiLines,
                        let ds = [d | d<-ds0,not (isInstance d),
				      include (defName d)],
                        not (null ds)]

fudgetIndex today hiLines indexName crossrefs synopses =
    on "!DOCTYPE html" $$
    on "HTML" $$
    title2 (manualName++" - "++indexName) ("¤ "++manualName++" ¤")
     (plain ("Created from the "++libraryName++" sources on "++today) $$
      wrap "H2" (plain indexName) $$
      indexSections hiLines) $$
     vcat [off "BODY",off "HTML"]
  where
    manualName = libraryName ++ " Reference Manual"

    indexSections ss =
      sectionList [n | Interface n _<-ss] $$
      vmap (fmtsection.ordersection) ss

    ordersection (Interface n s) =
        Interface n (nub types ++ rmdups (sort values))
      where (types,values) = partition isType s
            sort = sortBy by
            by d1 d2 = defName d1 `compare` defName d2

	    rmdups (x1:xs@(x2:_)) = if x1==x2
	                            then rmdups xs
				    else x1:rmdups xs
	    rmdups xs = xs

    fmtsection (Interface s ls) =
        wrap_id "H2" s (plain s) $$
	on "DL COMPACT" $$
       (vmap fmtGrp . grp . map (pairwith syn) $ ls) $$
	hcat [off "DL",on "HR"]
      where
        syn def = Map.lookup (defName def) synopses
	grp = moveMisc . map extr . groupBy eqSnd . sortBy cmpSnd
	  where
	    moveMisc (d@(Nothing,_):ds) = ds++[d]
	    moveMisc ds = ds
	    extr ((def,syn):ds) = (syn,def:map fst ds)
	    eqSnd = opSnd (==)
	    cmpSnd = opSnd compare
	    opSnd op (x1,y1) (x2,y2) = op y1 y2
        fmtGrp (syn,defs) =
	    fmtSyn (plain $ fromMaybe "Miscellaneous (the rest)" syn) $$
	    vmap fmtDef defs
	  where
	    fmtSyn s = on "DT"<>s<>text ":"
	    fmtDef d = on "DD"<>wrap "PRE" (ppTopDecl crossrefs d)

{-
    fmtsection (Interface s ls) =
        wrap "H2" (refpoint s):
	--on "P":
	on "DL":
	concatMap fmtline ls ++
	[off "DL",on "P",on "HR"]
      where
        fmtline d =
	    on "DT": on "CODE": ppTopDecl crossrefs d ++ [off "CODE"]++ synopsis
	  where
	    synopsis =
	      case Map.lookup (defName d) synopses of
	        Just s -> [on "DD"++s]
		_ -> []
-}

    sectionList ss =
      wrap "H2" (text "Sections") $$
      on "UL" $$
      vcat [on "LI"<>intref s|s<-ss] $$
      vcat [off "UL",tag "HR"]

fexamples (n@(this,_,_,_),ls) = (is,(n,ls'))
  where
    (is,ls') = fexamples' 0 [] ls
    fexamples' _ is [] = (is,[])
    fexamples' n is (cs:ls) =
       let (is',ls') = fexamples' n' is ls
           (kind,txt)  = takeWord cs
	   (n',is'',ls'') = if kind=="$FExample"
	                    then let imgname=this++show n'
			         in (n+1,
			             (imgname,txt):is',
				     ("\\image{"++imgname++".gif"++"}"):ls')
			    else (n,is',ls')
       in (is'', cs:ls'')

fexample2hs examples = unlines (
  "examples = ":
  ["  (\""++name++"\", stubF "++f name++") :" | (name,_) <- examples] ++
  ["  []",""] ++
  [f name ++ " = " ++ ex | (name,ex) <- examples])
  where
    f (c:cs) = toLower c:cs -- Let's hope this doesn't create any name clashes!

refpage names = [(n,page) | (page,ns) <- names, n<-page:ns]

writeDoc name d = writeFile' name (render' d)

render' = renderStyle style{lineLength=100,ribbonsPerLine=1}

writeFile' name s = putStrLn name >> writeFile name s
