#!/usr/local/bin/perl

sub Assert { 
    local($tst,$msg) = @_;
    $tst || die "$ARGV, line $.: $msg";
}

sub Warning {
    local($msg) = @_;
    print STDERR "Warning: $ARGV, line $.: $msg\n";
}

sub checkArg {
    local($k1,$k2,$arg) = @_;
    $k1 eq $k2 # && Assert($arg,"Missing argument to \$$k1");
}

sub iskey1 {
#    if(/^\$(\w+)\s*(\w+)?/) {
    if(/^\$(\w+)\s*([^\s]+)?/) {
	($1,$2);
    }
}

sub put {
    push(docbuf,@_);
}

sub putsect {
    &put("\\subsection*{@_}");
#    &put("");
}

sub putrest {
    &put($_);
}

sub writeIfDifferent {
    local($fname,@content) = @_;
    local($equal) = 0;
    if(open(oldf,$fname)) {
	$equal = 1;
	while(<oldf>) {
	    chop;
	    $equal = @content[$.-1] eq $_;
	    last if !$equal;
	}
	$equal = $equal && $. == @content;
	close(oldf);
    }
    if(!$equal) {
	&Assert(open(outf,">$fname"),"Cannot write $fname");
	print outf join("\n",@content),"\n";
	close(outf);
    }
}

sub procFudgetExample {
    local($fname,$fudget) = @_;
    local($pgmname) = "$fname-ex.hs";
    local($equal) = 0;
    local($title) = "FudgetExample";
    local(@pgm) = (
		   "module Main(main) where",
		   "import Fudgets",
		   "pos = Point 100 100",
		   "main = fudlogue(shellF [StoreName \"$title\", SetNormalHints pos,",
		   "	MoveWindow pos] None nullK (sepF 5 ($fudget)))");
    &writeIfDifferent($pgmname,@pgm);
    &put("\\epsfig{$fname.ps}");
}

sub procExample {
    local(@excode) = @_;
    &putsect("Example");
    &put("\\begin{example}");
    local($l);
    foreach $l (@excode) {
	$l =~ s/\\/{\\back}/g;	# should protect more...
	&put($l);
    }
    &put("\\end{example}");
}

sub procType {
    local($name) = @_;
    &put("\\begin{example}");
    &put(@types{$name});
    &put("\\end{example}");
}

sub getTypes {
    local($fname) = @_;
    if(!open(tf, "<$fname")) { 
	&Warning ("Can't find $fname.hi"); 
	return ();
    }
    local(@types);
    while(<tf>) {
	chop;
	s/\s*\{-#.*//;
	s/;$//;
	if(/(\w+)\s*::\s*/) {
	    @types{$1} = $';
	} elsif(/^(data|type)\s+(\w+)/) {
	    @types{$2} = $_;
	}
    }
    @types;
}

sub shiftArgType {
    local($argtypes,$del) = @_;
    local($type) = "";
    local($pno) = 0;
    local($p) = 0;
    while($c = substr($argtypes,$p,1)) {
	if(index("([",$c) !=-1) { $pno++; }
	if(index(")]",$c) !=-1) { $pno--; }
	if(!$pno && substr($argtypes,$p,length($del)) eq $del) {
	    last;
	}
	$type .= $c;
	$p++;
    }
    $type =~ s/^\s*//;
    $type =~ s/\s*$//;
    
    ($type,substr($argtypes,$p+length($del)));
}
	    
sub procDoc {
    local($fname, $argline,$ftype) = @_;
    local(@docbuf,@argbuf,$key,$arg,$inarg);
    local($argtypes,@atype,$atype,@iotype,$resulttype,@auxfnames);
    @auxfnames = split("[ \t]+",$argline);
    shift(@auxfnames);
    shift(@auxfnames);	      
    @docbuf = ();
    @argbuf = ();
    $argtypes = $ftype;
    $inarg = $argn = 0;
    while($argtypes) {
	($atype,$argtypes) = &shiftArgType($argtypes,"->");
	push(@atype,$atype);
    }
    while(<>) {
	chop;
	s/\$this/\\code{$fname}/;
	s/^\\xref{.*}$/$&./;
	($key,$arg) = &iskey1;
	if($key) {
	    s/^\$\w+\s*//;
	    if(&checkArg($key,"Arg",$arg)) {
		push(argbuf,$arg);
#		s/\w+\s*//;
		s/[^\s]+\s*//;
		if(!$inarg) { 
		    &putsect("Arguments");
		    &put("\\begin{description}");
		}
		&put("\\item[\\code{$arg ::\\ $atype[$inarg++]}]");
	    } else {
		if($inarg) {
		    &put("\\end{description}");
		    $inarg = 0;
		}
		if(index("Input|Output",$key)!=-1) {
		    $resulttype = @atype[@atype-1];
		    if($resulttype =~ s/\s*F\s+//) { 
			(@iotype{'Input'},@iotype{'Output'}) = 
			    &shiftArgType($resulttype," ");
		    }

		    if(@iotype{$key}) {
			&putsect("$key :: @iotype{$key}");
		    } else {
			&putsect($key);
		    }
		}
		elsif(index("Bugs|Description",$key)!=-1) { 
		    &putsect($key);
		} elsif($key eq "See") {
		    &putsect("See also");
		} elsif(index($key,"Example")!=-1) { 
		    &procExample($_);
		&procFudgetExample("$dumpdir/$fname",$_) if $key eq "FExample";
		    $_ = "";
		} elsif($key eq "Enddoc") { 
		    last;
		} elsif($key eq "Type") {
		    &procType($arg);
		    $_ = "";
		} else { 
		    &putsect($key);
		    &Warning("Undefined key: $key");
		}
	    }
	}
	&putrest;
    }
    local(@allnames,$section);
    if($#auxfnames >= 0) {
	$section = "$fname etc";
    } else {
	$section = $fname;
    }
    @allnames = ($fname,@auxfnames);
    @filebuf{$fname} = 
	join("\n",
	     "\\node $fname, _NEXT, _PREV, $upnode",
	     "\\section{$section}\n");
    foreach $auxf (@allnames) {
	@filebuf{$fname} = @filebuf{$fname} .
	    "\\findex{$auxf}\n"
    }
    @filebuf{$fname} = @filebuf{$fname} .
	join("\n",
	     "\\subsection*{Synopsis}",
	     "\\code{$fname " . join(" ",@argbuf) . "}",
	     "\\subsection*{Types}",
	     "\\begin{example}",
	     "\\begin{iftex}\n");
    local($auxf,$auxts,$auxt,@auxt);
    foreach $auxf (@allnames) {
	@auxt = ();
	$auxts = @types{$auxf};
	while($auxts) {
	    ($auxt,$auxts) = &shiftArgType($auxts,"->");
	    push(@auxt,$auxt);
	}
	@filebuf{$fname} = @filebuf{$fname} .
	    "$auxf :: \\mbox{" . join("} -> \\mbox{",@auxt) . "}\n";
    }
    @filebuf{$fname} = @filebuf{$fname} .
	join("\n",
	     "\\end{iftex}",
	     "\\begin{ifinfo}\n");
    foreach $auxf (@allnames) {
	@filebuf{$fname} = @filebuf{$fname} .
	    "$auxf :: @types{$auxf}\n";
    }
    @filebuf{$fname} = @filebuf{$fname} .
	join("\n",
	     "\\end{ifinfo}",
	     "\\end{example}",
		 @docbuf) . "\n";
}

{
    $typefile = shift;
    $dumpdir = shift;
    $upnode = shift;
    @types = &getTypes($typefile);
    while(<>) {

#	if($ARGV ne $oldargv) {
#	    $oldargv = $ARGV;
#	    $module = $ARGV;
#	    $module =~ s/\.m$//;
#	    @types = &getTypes($module);
#	}

	($key,$arg) = &iskey1;
	if(&checkArg($key,"Doc",$arg)) { 
	    $argline = $_;
	    chop($argline);
	    @funcs = split(/[ \t]+/,$argline);
	    $func = $funcs[1];
	    if($#funcs != 1) {
		$group{$func} = $argline;
	    }
	    &procDoc($func,$argline,@types{$func});
	    if (eof) {
		close(ARGV); # reset line number
	    }
	}
    }

# connect nodes

    @sortkeys = sort (keys %filebuf);
#    @sortkeys = keys %filebuf;
    foreach $this (@sortkeys) {
	if($prev) {
	    @filebuf{$prev} =~ s/_PREV/$prevprev/;
	    @filebuf{$prev} =~ s/_NEXT/$this/;
	}
	$prevprev = $prev;
	$prev = $this;
    }
    @filebuf{$prev} =~ s/_PREV/$prevprev/;
    @filebuf{$prev} =~ s/_NEXT/$this/;

# make menu and spit out

    print "\\begin{menu}\n";
    foreach $this (@sortkeys) {
	if($group{$this} eq "") {
	    print "* $this::\n";
	} else {
	    @funcs = split(/[ \t]+/,$group{$this});
	    shift(@funcs);
	    $groupname = shift(@funcs);
	    print "* $groupname::\n";
	    foreach $func (@funcs) {
		print "* $func: $groupname\n";
	    }
	}
    }
    print "\\end{menu}\n";
    foreach $this (@sortkeys) {
	print @filebuf{$this};
    }
}

