#include <stdio.h>
#include <errno.h>
#define XLIB_ILLEGAL_ACCESS	/* to access Display struct in Ultrix */
#include <X11/Xlib.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <unistd.h>
#include "heap.h"
#include "sockets.h"
#include "dialog.h"
#include "timeq.h"
#include "xdecode.h"
#include "xencode.h" /* mknstring */
#include "asyncinput.h"

extern int hostName();

#ifdef _HPUX_SOURCE
#define DTABLESIZE() sysconf(_SC_OPEN_MAX)
#else
#define DTABLESIZE() getdtablesize()
#endif

#define MOTIONCOMPRESS 1

static PTR mkfailure PROTO ((int,char *));

/* The following two functions are modified copies of mkerresp and filerror
   in lib/dialog_c.c, which it would be better to restructure to avoid code
   duplication. */

static PTR
mkfailure(k, msg)
int k;
char *msg;
{
  if(debug) fprintf(stderr, "mkerrresp(%d,\"%s\")\n",k,msg);
    return mkconstr(Failure, 1, mkconstr(k, 1, mkestring(msg)));
}

PTR
mkfileerror(r, fname, op)
int r;
char *fname, *op;
{
    char buf[10000];

    if (fname)
	sprintf(buf, "%s: %s: %s", op, sys_errlist[errno], fname);
    else
	sprintf(buf, "%s: %s", op, sys_errlist[errno]);
    return mkfailure(r, buf);
}

#define ERR(s,blaha) return mkfileerror(OtherError,NULL,s)

#define BUFSIZE 2000
#define WORDSPERCHAR 5

FILE* getfilep();

typedef struct {
  PTR p;
  int cno;
} Aitable;

static Aitable* aitable;
static int aitablesize;

static fd_set asyncfds,asyncwritefds;

#if MOTIONCOMPRESS
static int motioncompress = 1;
#endif

static int getdisplayfd(d)
Display *d;
{
  return d->fd;
}

static void initaitable() {
  if(aitable) return;
  aitablesize = DTABLESIZE();

  if(debug) fprintf(stderr,"aitablesize %d\n",aitablesize);
  if(!(aitable = (Aitable *)calloc(aitablesize,sizeof(Aitable)))) {
    fprintf(stderr,"Cannot allocate aitable: Out of memory\n");
    exit(1);
  }
  FD_ZERO(&asyncfds);
  FD_ZERO(&asyncwritefds);
}

void asyncrequest(p) /* Select */
PTR p;
{
  int fd;
  PTR l;

  p = evaluate(p);
  initaitable();
  for(fd=0;fd<aitablesize;fd++) {
    aitable[fd].p = NULL;
    aitable[fd].cno = -1;
  }

  FD_ZERO(&asyncfds);
  FD_ZERO(&asyncwritefds);
  if(debug) fprintf(stderr,"Select: ");
  for(l=p; !ISNIL(l); l = ETLOF(l)) {
    PTR d;
    int cno;
    PUSHPTR(l);
    d = EHDOF(l);
    cno = getcno(d);
    if(debug) fprintf(stderr,"cno=%d ",cno),fflush(stderr);
    if(cno != TimerDe) {
      switch(cno) {
      case LSocketDe:
      case SocketDe:
      case OutputSocketDe:
	{ FILE *fp = (FILE*)INTOF(EARG1(d));
	  REPLAY_ONE_AND_SKIP(fd)
	    fd = fileno(fp);
	  RECORD_ONE(fd);
	  if(debug) fprintf(stderr,"{L}SocketDe %d, ",fd);
	} break;
      case DisplayDe:
	{ Display *dp = (Display *)INTOF(EARG1(d));
	  REPLAY_ONE_AND_SKIP(fd)
	    fd = getdisplayfd(dp);
	  RECORD_ONE(fd);
	  if(debug) fprintf(stderr,"DisplayDe %d, ",fd);
	} break;
      default:
	fprintf(stderr,"Unknown descriptor: %d\n",cno);
	exit(1);
      }
      if(aitable[fd].cno==SocketDe && cno==OutputSocketDe
	 || aitable[fd].cno==OutputSocketDe && cno==SocketDe) {
	cno=aitable[fd].cno=InputOutputSocketDe;
	if(debug) fprintf(stderr,"fd %d is read&write, ",fd);
      }
      else
	aitable[fd].cno = cno;
      aitable[fd].p = (PTR)INTOF(getpart(d,0,1));
      if(cno!=OutputSocketDe)
	FD_SET(fd,&asyncfds);
      switch(cno) {
      case OutputSocketDe:
      case InputOutputSocketDe:
	FD_SET(fd,&asyncwritefds);
	if(debug) fprintf(stderr,"Adding fd %d to asyncwritefds\n",fd);
      }
    }
    POPPTR(l);
  }
  if(debug) fprintf(stderr,"\n");
}

PTR mkEventconstr(fd)
int fd;
{
  Display *disp;
  XEvent xevent;

#ifdef PROFILE
  if(replay) {
    REPLAY(disp);
    REPLAY(xevent);
  } else {
#endif
    disp = (Display *)aitable[fd].p;
    XNextEvent(disp,&xevent);

#if MOTIONCOMPRESS
    if (xevent.type == MotionNotify && motioncompress) {
      while(XEventsQueued(disp, QueuedAlready) > 0) {
	XEvent ev;
	XPeekEvent(disp, &ev);
	if (ev.type != MotionNotify)
	  break;
	XNextEvent(disp, &xevent);
	/* compr++; */
      }
    }
#endif
#ifdef PROFILE
  }
  if(record) {
    RECORD(disp);
    RECORD(xevent);
  }
#endif

  return mkconstr1(EXEvent,mkpair(mkint((Int)xevent.xany.window),
				  encodeEvent(&xevent)));   /* encodeEvent knows about two-pass profiling */
}

#if 0
PTR mkestring2 PROTO((char *,int));

PTR
mkestring2(as,l)
char *as;
int l;
{
    unsigned char *s = (unsigned char *)as;
    PTR p;

    for(s += l, p = mknil(); l > 0; l--) {
      if (NEEDGC) {
	PUSHPTR(p);
	DOGC;
	POPPTR(p);
	if(debug)
	  fprintf(stderr,"GC in mkestring2 in asyncinput.c\n");
      }
      p = mkcons(mkchar(*--s), p);
    }
    return p;
}
#endif

PTR 
getnextasync(void)
{
  int n,cno;
  int fd;
  PTR event;
  fd_set readfds,writefds;
  struct timeval now, *timeout;
  initaitable();
  event = NULL;

  for(fd=0;fd<aitablesize;fd++)
    if(aitable[fd].cno==DisplayDe) {
      int bool;
      REPLAY_ONE_AND_SKIP(bool)
	bool = XEventsQueued((Display *)aitable[fd].p,QueuedAfterFlush);
      RECORD_ONE(bool);
      if(bool) {
	if(debug) fprintf(stderr,"getnextasync: XEventsQueued %d\n",fd);
	event = mkEventconstr(fd);  /* Knows about two-pass profiling */
	if(debug) fprintf(stderr,"event cno: %d\n",getcno(SNDOF(getpart(event,0,1))));
	break;
      }
    }
  if(!event) {
    readfds = asyncfds;
    writefds = asyncwritefds;
  WHEN_NOT_REPLAY_START
    if (!QISEMPTY()) {
      unsigned long nowms;
      long tleft;
      gettimeofday(&now, (struct timezone *)0);
      nowms = TIMETOMS(&now);
      tleft = QTIME() - nowms;
      if (debug) fprintf(stderr, "tleft=%ld\n", tleft);
      if (tleft < 0) tleft = 0;
      now.tv_usec = (tleft % 1000) * 1000;
      now.tv_sec = tleft / 1000;
      timeout = &now;
      if (debug) fprintf(stderr, "timeout in event needed %ld %ld\n", (long)now.tv_sec, (long)now.tv_usec);
    } else
      timeout = NULL;
  WHEN_NOT_REPLAY_END
    if(debug) fprintf(stderr,"getnextasync: select...");
#ifdef PROFILE
    if(replay) {
      REPLAY(n);
      REPLAY(errno);
      REPLAY(readfds);
      REPLAY(writefds);
    } else 
#endif
      do {
	n = select(aitablesize,&readfds,&writefds,NULL,timeout);
	if(debug) fprintf(stderr," = %d\n",n);
      } while(n == -1 && errno == EINTR);
#ifdef PROFILE
    if(record) {
      RECORD(n);
      RECORD(errno);
      RECORD(readfds);
      RECORD(writefds);
    }
#endif
    
    if(n == -1) { ERR("select",NULL); }
    if(n > 0) {
      REPLAY_ONE_AND_SKIP(fd)
	for(fd=0;!FD_ISSET(fd,&readfds) && !FD_ISSET(fd,&writefds); fd++)
	  ;
      RECORD_ONE(fd);
      cno=aitable[fd].cno;
      if(debug) fprintf(stderr,"fd %d, cno %d, is ready for action\n",fd,cno);
      switch(cno) {
      case LSocketDe: {
	struct sockaddr addr;
	char peername[1000];
	int addrlen=sizeof(addr);
	int sfd;
	FILE *sf;
	
	REPLAY_ONE_AND_SKIP(sf) {
	  sfd = accept(fd,&addr,&addrlen);
	  if(sfd<0) { ERR("accept",NULL); }
	  sf = getfilep(sfd,"r+");
	  if(hostName(&addr,peername)) {ERR("hostName",NULL); }
	  if(!sf) { ERR("fdopen",NULL); }
	  if(debug) fprintf(stderr,"SocketAccepted %d %lx \"%s\"\n",fd,(Int)sf,peername);
	}
	RECORD_ONE(sf);
	event = mkconstr(ESocketAccepted,2,mkint((Int)sf),mkestring(peername)); /* mkestring knows about two-pass profiling */
	break; }
      case OutputSocketDe:
      case InputOutputSocketDe:
      case SocketDe:
	if(cno==OutputSocketDe ||
	   cno==InputOutputSocketDe && FD_ISSET(fd,&writefds)) {
	  /* If both read and write are possible, write has priority. */
	  cno=OutputSocketDe;
	  event=mkconstr0(ESocketWritable);
	}
	else {
	  char buf[BUFSIZE];
	  int got;

	  REPLAY_ONE_AND_SKIP(got)
	    got = read(fd,buf,(sizeof buf - 1));
	  RECORD_ONE(got);
	  if(got<0) { ERR("read",NULL); }
	  if(debug) {
	    fprintf(stderr,"SocketRead %d bytes from descr %d: \"%s\"\n",
		    got,fd,debug>1 ? buf : "...");
	    if(got==sizeof buf-1) fprintf(stderr,"Buffer was full\n");
	    /* EOF -> empty string */
	  }
	  cno=SocketDe;
	  event = mkconstr1(ESocketRead,mknstring(buf,got)); /* mknstring knows about two-pass profiling */
	}
	break;
      case DisplayDe:
	event = mkEventconstr(fd);
	break;
      default:
	fprintf(stderr,"Unexpected fd in select: %d\n",fd);
	exit(1);
      }
    } else { /* n == 0, timeout */
      struct timeq e;
      REPLAY_ONE_AND_SKIP(e.msg)
	e = removetimeq();
      RECORD_ONE(e.msg);
      if(debug) fprintf(stderr, "TimerAlarm %d\n",e.msg);
      return
	mkconstr1(AsyncInput,mkpair(mkconstr1(TimerDe,mkint(e.msg)),
				    mkconstr0(ETimerAlarm)));
    }
  }

  REPLAY_ONE_AND_SKIP(aitable[fd])
    ;
  RECORD_ONE(aitable[fd]);
  if(debug) fprintf(stderr,"getnextasync returning fd %d, cno %d\n",fd,cno);
  return
    mkconstr1(AsyncInput,
	      mkpair(mkconstr1(cno,mkint((Int)aitable[fd].p)),
		     event));
}
