/*
 * (c) Copyright 1991 by Panagiotis Tsirigotis
 * Read the file COPYRIGHT for more details.
 * 
 */
/*
 * $Header: /Users/magnus/cvs/Fudgets/xlib/socketlib/ne.c,v 1.2 2007/02/25 14:51:59 hallgren Exp $
 */

/*
 * This file contains the functions of the socket library that
 * either succeed or terminate the program
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>
#include <string.h>

#include "defs.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#define PRIVATE			static

#if 1
#define ERROR_MESSAGE(val) strerror(val)
#else
#define ERROR_MESSAGE(val) ((val>=sys_nerr) ? "unknown error"	\
					    : sys_errlist[ val ] )
#endif

#define FAIL(call) fprintf( stderr, "%s failed: errno = %d (%s)\n", \
			    call, errno, ERROR_MESSAGE(errno))

/*
extern int sys_nerr ;
extern char *sys_errlist[] ;
extern int errno ;
*/

/* VARARGS1 */
PRIVATE void fail(char *call_name, ... )
{
	va_list ap ;
	char *format ;

	FAIL( call_name ) ;
	va_start( ap, call_name ) ;
	format = va_arg( ap, char * ) ;
	vfprintf( stderr, format, ap ) ;
	va_end( ap ) ;
	exit( 1 ) ;
	/* NOTREACHED */
}


#define TYPE_NAME_LEN					10

/*
 * Returns a string with the socket type (e.g. "SOCK_STREAM" for SOCK_STREAM)
 */
PRIVATE char *socket_type_name( type )
	int type ;
{
	static char type_name[ TYPE_NAME_LEN ] ;

	if ( type == SOCK_STREAM )
		return( "SOCK_STREAM" ) ;
	else if ( type == SOCK_DGRAM )
		return( "SOCK_DGRAM" ) ;
	else
	{
		(void) sprintf( type_name, "%d", type ) ;
		return( type_name ) ;
	}
}


/*
 * This function prints a description of the resolver error
 * and terminates the process
 */
PRIVATE void resolver_error( call, host )
	char *call ;
	char *host ;
{
	fprintf( stderr, "%s: resolver error. ", call ) ;

#ifdef HAS_H_ERRNO
	switch ( errno )
	{
		case HOST_NOT_FOUND:
			fprintf( stderr, "Host not found. " ) ;
			break ;
		
		case TRY_AGAIN:
			fprintf( stderr, "Got a TRY_AGAIN after %d times. ", MAX_TRIES ) ;
			break ;
		
		case NO_RECOVERY:
			fprintf( stderr, "Non-recoverable error. " ) ;
			break ;
		
		case NO_DATA:
			fprintf( stderr,
						"Valid name, but no data record of requested type. " ) ;
			break ;

#if NO_ADDRESS != NO_DATA
		case NO_ADDRESS:
			fprintf( stderr, "No address. " ) ;
			break ;
#endif
		
		default:
			fprintf( stderr, "UNKNOWN h_errno VALUE (h_errno=%d). ", errno ) ;
			break ;
	}
#endif 	/* HAS_H_ERRNO */
	fprintf( stderr, "Host = %s\n", ( host ) ? host : "LOCAL" ) ;
	exit( 1 ) ;
}


void in_address_ne( host, port, address )
	char *host ;
	short port ;
	struct sockaddr_in *address ;
{
	if ( in_address( host, port, address ) == RESOLVER_ERROR )
		resolver_error( "in_address_ne", host ) ;
}


int in_bind_ne(short port, int type )
{
	char num[ 10 ] ;
	char *type_name ;
	int sd = in_bind( port, type ) ;

	if ( sd < 0 )
		switch ( sd )
		{
			case SYSCALL_ERROR:
				fail( "in_bind_ne: bind", "port = %d, type = %s\n", 
					port, socket_type_name( type ) ) ;
			/* NOTREACHED */
			
			case RESOLVER_ERROR:
				resolver_error( "in_bind_ne", NULL ) ;
			/* NOTREACHED */

			default:
				fprintf( stderr,
						"in_bind_ne: in_bind RETURNED BAD VALUE: %d\n", sd ) ;
				exit( 1 ) ;
		}
	return( sd ) ;
}


in_connect_ne( host, port, type )
	char *host ;
	short port ;
	int type ;
{
	int sd = in_connect( host, port, type ) ;

	if ( sd < 0 )
		switch ( sd )
		{
			case SYSCALL_ERROR:
				fail( "in_connect_ne: connect", "host = %s, port = %d, type = %s\n",
					host, port, socket_type_name( type ) ) ;
				/* NOTREACHED */
			
			case RESOLVER_ERROR:
				resolver_error( "in_connect_ne", host ) ;
				/* NOTREACHED */
			
			default:
				fprintf( stderr,
						"in_connect_ne: in_connect RETURNED BAD VALUE: %d\n", sd ) ;
				exit( 1 ) ;
			}
	return( sd ) ;
}


int un_bind_ne( path, type )
	char *path ;
	int type ;
{
	int sd = un_bind( path, type ) ;
	
	if ( sd == SYSCALL_ERROR )
		fail( "un_bind_ne: bind", "path = %s, type = %s\n",
											path, socket_type_name( type ) ) ;
	return( sd ) ;
}


int un_connect_ne( path, type )
	char *path ;
	int type ;
{
	int sd = un_connect( path, type ) ;

	if ( sd == SYSCALL_ERROR )
		fail( "un_connect_ne: connect", "path = %s, type = %s\n",
											path, socket_type_name( type ) ) ;
	return( sd ) ;
}


void un_close_ne( sd )
	int sd ;
{
	if ( un_close( sd ) == SYSCALL_ERROR )
		fail( "un_close_ne", "sd = %d\n", sd ) ;
}

