.\" (c) Copyright 1991 by Panagiotis Tsirigotis
.\" Read the file COPYRIGHT for more details.
.\"
.\" $Header: /Users/magnus/cvs/Fudgets/xlib/socketlib/socket.man,v 1.1.1.1 1996/03/10 16:56:13 magnus Exp $
.TH LIBSOCKET 3PT "19 March 1991"
.SH NAME
in_address, in_address_ne, in_bind, in_bind_ne, in_bind_opt, in_connect, in_connect_ne, in_connect_opt, un_address, un_bind, un_bind_ne, un_connect, un_connect_ne, un_close, un_close_ne \- socket support library
.SH SYNOPSIS
.LP
.nf
.ft B
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
.LP
.ft B
int in_address( host, port, address )
char *host ;
short port ;
struct sockaddr_in *address ;
.LP
.ft B
void in_address_ne( host, port, address )
char *host ;
short port ;
struct sockaddr_in *address ;
.LP
.ft B
int in_bind( port, type )
short port ;
int type ;
.LP
.ft B
int in_bind_ne( port, type )
short port ;
int type ;
.LP
.ft B
int in_bind_opt( port, type, [ option, value, \.\.\. ], NULL )
short port ;
int type ;
int option ;
int value ;
.LP
.ft B
int in_connect( host, port, type )
char *host ;
short port ;
int type ;
.LP
.ft B
int in_connect_ne( host, port, type )
char *host ;
short port ;
int type ;
.LP
.ft B
int in_connect_opt( host, port, type, [ option, value, \.\.\. ], NULL )
char *host ;
short port ;
int type ;
int option ;
int value ;
.LP
.ft B
void un_address( path, address )
char *path ;
struct sockaddr_un *address ;
.LP
.ft B
int un_bind( path, type )
char *path ;
int type ;
.LP
.ft B
int un_bind_ne( path, type )
char *path ;
int type ;
.LP
.ft B
int un_connect( path, type )
char *path ;
int type ;
.LP
.ft B
int un_connect_ne( path, type )
char *path ;
int type ;
.LP
.ft B
int un_close( sd )
int sd ;
.LP
.ft B
void un_close( sd )
int sd ;
.SH DESCRIPTION
.LP
The functions that have a \fIin_\fR
prefix support \fB Internet \fR domain sockets while those with a
\fIun_\fR prefix support \fBUnix\fR domain sockets.
.LP
Functions with a \fI_ne\fR suffix act the same as the ones without
the suffix, but they never return an error value; instead they
print (at stderr) an error message with the function name and
the type of error and then they terminate the process.
These functions have the same type as the ones without the
\fI_ne\fR suffix only when the return value is
a file descriptor; otherwise the function type is \fIvoid\fR.
There will be no other description of these functions.
.LP
.B in_address()
forms an Internet socket address that can be used in a socket system call
(e.g. \fIbind\fR).
The \fIhost\fR argument can be either a symbolic host name
(like foo.colorado.edu) or a numeric address (like 128.138.211.2).
If \fIhost\fR is
.SM NULL
then the local host (as returned by \fIgethostname(2)\fR) is assumed.
If you don't use this function, you need not include 
.I <netinet/in.h>.
.LP
.B in_bind()
binds creates a socket and binds it to the address
(\fBINADDR_ANY\fR,\fIport\fR)
(the \fIport\fR is provided by the caller). The \fItype\fR
argument determines the type of the socket. It may be \fBSOCK_STREAM\fR
or \fBSOCK_DGRAM.\fR
.B in_bind_opt()
does the same job as 
.B in_bind()
but it also sets \fIsocket level\fR options. The options supported are:
.B "SO_DEBUG, SO_REUSEADDR, SO_KEEPALIVE, SO_DONTROUTE, SO_BROADCAST,"
.B SO_OOBONLINE.
The \fIvalue\fR of the option should be either 0 or 1.
The \fI(option,value)\fR list is terminated by a
.SM NULL
argument.
.LP
.B in_connect()
creates a socket and connects it to the address \fI(host,port)\fR.
The \fIhost\fR argument can be either a symbolic host name
(like foo.colorado.edu) or a numeric address (like 128.138.211.2).
If \fIhost\fR is
.SM NULL
then the local host (as returned by \fIgethostname(2)\fR) is assumed.
The \fItype\fR argument determines the type of the socket. It may be
\fBSOCK_STREAM\fR or \fBSOCK_DGRAM.\fR
.B in_connect_opt()
does the same job as
.B in_connect()
but it also sets \fIsocket level\fR options.
The supported options and the form of the option list are the same as that of
.B in_bind_opt().
.LP
.B un_address()
forms an Unix socket address that can be used in a socket system call
(e.g. \fIbind\fR). If you don't use this function, you need not include 
.I <sys/un.h>.
.LP
.B un_bind()
creates a socket and binds it to the specified \fIpath\fR.
It also records the \fIpath\fR (this is used by \fBun_close()\fR).
The recording of path names will stop if there is no more memory
(i.e. \fImalloc(\|)\fR fails).
\fItype\fR determines is the type of the socket. It may be
\fBSOCK_STREAM\fR or \fBSOCK_DGRAM.\fR
.LP
.B un_connect()
creates a socket and connects it to the specified \fIpath\fR.
\fItype\fR determines is the type of the socket. It may be
\fBSOCK_STREAM\fR or \fBSOCK_DGRAM.\fR
.LP
.B un_close()
closes a Unix domain socket and unlinks the socket file. The socket
must be created by
.B un_bind()
so the path of the socket is recorded.
.LP
.SH RETURN VALUES
.LP
On success,
.B in_bind()
and
.B in_connect()
return a file descriptor. On failure, they return, 
.B -1
in case of a system call error, or
.B -2
in case of a resolver error.
.B errno
is set to indicate the error (in case of a resolver error,
.B errno
has the value of
.B h_errno.
.LP
On success,
.B in_bind_opt()
and
.B in_connect_opt()
return a file descriptor. On failure, they return
.B -1
in case of a system call error,
.B -2
in case of a resolver error, or
.B -3
in case of an option error.
.B errno
is set set to indicate the error (in case of a resolver error,
.B errno
has the value of
.B h_errno
).
.LP
On success,
.B in_address()
returns
.B 0.
On failure, it returns
.B -2
and sets
.B errno
to the value of
.B h_errno.
.LP
On success,
.B un_bind()
and
.B un_connect()
return a file descriptor.
On failure, they return
.B -1
and
.B errno
is set to indicate the error.
.LP
On success,
.B un_close()
returns
.B 0.
On failure, it returns
.B -1
and
.B errno
is set to indicate the error.

.SH BUGS
.LP
When a function returns -2 (resolver error),
.B errno
is set to the value of
.B h_errno
only is the resolver routines set
.B h_errno.
.LP
.B in_bind_opt()
and
.B in_connect_opt()
can only set binary-value options. Furthermore, they can only set
socket level options.
