/*
 * (c) Copyright 1991 by Panagiotis Tsirigotis
 * Read the file COPYRIGHT for more details.
 * 
 */

/*
 * UNIX domain socket support library
 *
 *  INTERFACE:
 *
 *		int un_address		: form a UNIX domain socket address
 *		int un_bind			: bind a socket to a UNIX address
 *		int un_connect		: connect to a UNIX address
 *		int un_close		: close a socket and unlink the pathname of the
 * 							  socket
 *
 */

/*
 * $Header: /Users/magnus/cvs/Fudgets/xlib/socketlib/unix.c,v 1.3 2009/12/17 15:37:53 hallgren Exp $
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <errno.h>
#include <string.h>

#define NO		0
#define YES		1

/*
 * Close a file descriptor without destroying errno
 */
#define CLOSE( sd, t )				t = errno ;						\
											(void) close( sd ) ;			\
											errno = t


static int first_bind = YES ;
static int paths_saved ;			/* initialized at first_bind */
static char **paths ;



/*
 * Creates a Unix domain socket address
 */
void un_address( pathname, address )
	char *pathname ;
	struct sockaddr_un *address ;
{
	address->sun_family = AF_UNIX ;
	(void) strcpy( address->sun_path, pathname ) ;
}


int un_bind( path, type )
	char *path ;
	int type ;
{
	int sd ;
	struct sockaddr_un addr ;
	int errno_save ;

	if ( ( sd = socket( AF_UNIX, type, 0 ) ) == -1 )
		return( -1 ) ;

	un_address( path, &addr ) ;

	if ( bind( sd, (struct sockaddr *) &addr,
					sizeof( addr.sun_family ) + strlen( addr.sun_path ) ) == -1 )
	{
		CLOSE( sd, errno_save ) ;
		return( -1 ) ;
	}

	if ( first_bind )
	{
		int i ;
		int n_descriptors = getdtablesize() ;

		paths = (char **) malloc( n_descriptors * sizeof( char * ) ) ;
		if ( paths != NULL )
		{
			for ( i = 0 ; i < n_descriptors ; i++ )
				paths[ i ] = NULL ;
			paths_saved = YES ;
		}
		else
			paths_saved = NO ;
	}
	if ( paths_saved )
	{
		if ( paths[ sd ] != NULL )
			(void) free( paths[ sd ] ) ;
		paths[ sd ] = malloc( strlen( addr.sun_path ) + 1 ) ;
		if ( paths[ sd ] != NULL )
			(void) strcpy( paths[ sd ], addr.sun_path ) ;
		else
			paths_saved = NO ;
	}

	return( sd ) ;
}


int un_connect( path, type )
	char *path ;
	int type ;
{
	int sd ;
	struct sockaddr_un addr ;
	int errno_save ;

	if ( ( sd = socket( AF_UNIX, type, 0 ) ) == -1 )
		return( -1 ) ;

	un_address( path, &addr ) ;

	if ( connect( sd, (struct sockaddr *) &addr,
					sizeof( addr.sun_family ) + strlen( addr.sun_path ) ) == -1 )
	{
		CLOSE( sd, errno_save ) ;
		return( -1 ) ;
	}
	return( sd ) ;
}


int un_close( sd )
int sd ;
{
	if ( paths[ sd ] != NULL )
	{
		(void) unlink( paths[ sd ] ) ;
		(void) free( paths[ sd ] ) ;
		paths[ sd ] = NULL ;
	}
	return( close( sd ) ) ;
}

