#include <sys/types.h>
#include <sys/socket.h>
#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include "dialog.h"
#include "heap.h"
#include "sockets.h"
#include "xdecode.h" 
#include "xencode.h" /* mknstring */
#include "timeq.h"
#include "asyncinput.h"
#include "doxcall.h"

#if defined(SOLARIS) || defined(__NetBSD__) || defined(__FreeBSD__) || defined(__linux__)
#include <dlfcn.h>
#define SUPPORT_DYNLINK 1
#if defined(SOLARIS)
#define DLOPEN_MODE RTLD_NOW
#elif defined(RTLD_LAZY)
#define DLOPEN_MODE RTLD_LAZY
#else
#define DLOPEN_MODE 1
#endif
#endif

extern void fileerror();
extern int outstring();
extern void mkresponse();
extern int in_bind();
extern int in_connect();

#define mkSuccess() mkconstr0(RSuccess)

extern int debug;
static PTR socketrequest PROTO ((PTR));

#define getport(p) INTOF(p)

typedef Int Port,Socket;

#define ERR(s) return mkfileerror(OtherError,NULL,(s))

#define mkSocket(f) mkint((Int)f)
#define mkLSocket(f) mkint((Int)f)

#define mkSRSocket(f) mkconstr1(SRSocket,mkSocket(f))
#define mkSRLSocket(f) mkconstr1(SRLSocket,mkLSocket(f))

static PTR mkSocketResponse(PTR p)
{
  return mkconstr1(SocketResponse,p);
}

FILE* getfilep(s,type)
Socket s;
char *type;
{
  FILE *sf = fdopen(s,type);
  if (!sf && errno == EMFILE) {
    DOGC;
    sf = fdopen(s, type);
  }
#if 0
  if(sf)
    addgcfile(sf, (DIR*)0);
#endif
  return sf;
}

static PTR startProcess PROTO((char *,int doIn,int doOut,int doErr));

static PTR
startProcess(cmd,doIn,doOut,doErr)
char *cmd;
int doIn,doOut,doErr;
{
  int inPipe[2],outPipe[2],errPipe[2];

  if(doIn) pipe(inPipe);
  if(doOut) pipe(outPipe);
  if(doErr) pipe(errPipe);
  
  switch(fork()) {
  case -1:
    ERR("fork");
    break;
  case 0:
    /* child process */
    if(doIn)  close(0), dup(inPipe[0]),
              close(inPipe[0]),  close(inPipe[1]);
    if(doOut) close(1), dup(outPipe[1]),
              close(outPipe[0]), close(outPipe[1]);
    if(doErr) close(2), dup(errPipe[1]),
              close(errPipe[0]), close(errPipe[1]);
    execl("/bin/sh","sh","-c",cmd,NULL);
    /* exec failed. Should generate some error message!! */
    _exit(1); 
  default:
    /* parent process */
    { PTR inS,outS,errS;
      FILE *f;

      if(doIn) {
	close(inPipe[0]);
	f=getfilep(inPipe[1],"w");
	if(debug) fprintf(stderr,"Fd in=%d\n",inPipe[0]);
	if(!f) { ERR("fdopen"); break; }
	inS=mkSome(mkSocket(f));
      }
      else inS=mkNone();
      if(doOut) {
	close(outPipe[1]);
	f=getfilep(outPipe[0],"r");
	if(debug) fprintf(stderr,"Fd out=%d\n",outPipe[1]);
	if(!f) { ERR("fdopen"); break; }
	outS=mkSome(mkSocket(f));
      }
      else outS=mkNone();
      if(doErr) {
	close(errPipe[1]);
	f=getfilep(errPipe[0],"r");
	if(debug) fprintf(stderr,"Fd err=%d\n",errPipe[1]);
	if(!f) { ERR("fdopen"); break; }
	errS=mkSome(mkSocket(f));
      }
      else errS=mkNone();
      return mkSocketResponse(mkconstr(SRProcessSockets,3,inS,outS,errS));
    }
  }
  return 0; /* Should be unreachable */
}

static PTR socketrequest(PTR p)
{
  PTR resp;
  p = evaluate(p);
  switch(getcno(p)) {
  case SRqOpenLSocket: /* Port */
    {
      Port port;
      Socket s;
      FILE *sf;

      port = getport(EARG1(p));
      REPLAY_ONE_AND_SKIP(sf) {
	s = port ? in_bind(port,SOCK_STREAM)
	         : socket(AF_INET,SOCK_STREAM,0);
                 /* port==0 ==> pick a free port number */

	if(debug) fprintf(stderr,"OpenLSocket %lx %d\n",(Int)port,SOCK_STREAM);
	if(s<0) { ERR("in_bind"); break; }
	if (listen(s,5)) { ERR("listen"); break; }
	sf = getfilep(s,"r+");
      }
      RECORD_ONE(sf);
      if(!sf) { ERR("fdopen"); break; }
      if(debug) fprintf(stderr,"LSocket %lx\n",(Int)sf);
      resp=mkSocketResponse(mkSRLSocket(sf)); break; 
    }
  case SRqOpenSocket: /* Host Port */
    {
      char host[10000];
      Port port;
      FILE *sf;
      Socket s;

      p = evalArgs(p,2);
      REPLAY_ONE_AND_SKIP(port)
	port = getport(getpart(p,1,2));
      RECORD_ONE(port);
      evalstring(getpart(p,0,2), host, sizeof host);
      if(debug) fprintf(stderr,"OpenSocket \"%s\" %ld %d\n",host,(Int)port,SOCK_STREAM);
      REPLAY_ONE_AND_SKIP(sf) {
	s = in_connect(host,port,SOCK_STREAM);
	if(s<0) { ERR("in_connect"); break; }
	sf = getfilep(s,"r+");
      }
      RECORD_ONE(sf);
      if(!sf) { ERR("fdopen"); break; }
      if(debug) fprintf(stderr,"Socket %lx\n",(Int)sf);
      resp=mkSocketResponse(mkSRSocket(sf)); break;
    }
  case SRqWriteSocket: /* Socket String */
    {
      FILE *sf;
      int bool;
      p = evalArgs(p,2);
      sf = (FILE *)INTOF(getpart(p,0,2));
      
      if(debug) fprintf(stderr,"WriteSocket %lx ...\n",(Int)sf);
      REPLAY_ONE_AND_SKIP(bool)
      bool = outstring(sf,getpart(p,1,2)) < 0;
      RECORD_ONE(bool);
      if(bool) { ERR("WriteSocket[outstring]"); break; }
      REPLAY_ONE_AND_SKIP(bool)
	bool = fflush(sf);
      RECORD_ONE(bool);
      if(bool) { ERR("WriteSocket[fflush]"); break; }
      resp=mkSuccess();
      break;
    }
  case SRqCloseSocket: /* Socket */
  case SRqCloseLSocket: /* LSocket */
    {
      FILE *sf = (FILE *)INTOF(EARG1(p));
      if(debug) fprintf(stderr,"Close{L}Socket %lx\n",(Int)sf);
      WHEN_NOT_REPLAY(fclose(sf));
      resp=mkSuccess();
      break;
    }
  case SRqGetStdinSocket: /* */
    resp=mkSocketResponse(mkSRSocket(stdin));
    break;
  case SRqGetStdoutSocket: /* */
    resp=mkSocketResponse(mkSRSocket(stdout));
    break;
  case SRqCreateTimer: /*  Int Int -- interval first */
    {
      static int timerno;
      int interval;
      int first;

      p = evalArgs(p,2);
      interval = INTOF(getpart(p,0,2));
      first = INTOF(getpart(p,1,2));
      WHEN_NOT_REPLAY(entertimeq(mktimeq(first, interval, timerno)));
      if (debug) fprintf(stderr, "CreateTimer(%d,%d) = %d\n", interval, first,timerno);
      resp=mkSocketResponse(mkconstr1(SRTimer,mkint(timerno)));
      timerno++;
      break;
    }
  case SRqDestroyTimer: /* Timer */
    {
      int tno = INTOF(EARG1(p));
      WHEN_NOT_REPLAY(cleartimeq(tno));
      if(debug) fprintf(stderr,"DestroyTimer %d\n",tno);
      resp=mkSuccess();
      break;
    }
  case SRqGetLSocketName: /* LSocket */
  case SRqGetSocketName: /* Socket */
    {
      FILE *sf;
      int s;
      struct sockaddr sa;
      int bool,len = sizeof sa;

      sf = (FILE *)INTOF(EARG1(p));
      REPLAY_ONE_AND_SKIP(bool) {
	s = fileno(sf);
	if(debug) fprintf(stderr,"GetLSocketName %lx\n",(Int)sf);
	bool = getsockname(s,&sa,&len);
      }
      RECORD_ONE(bool);
      if(bool)
	ERR("GetLSocketName");
      else
	resp=mkconstr1(Str,mknstring(sa.sa_data,len));  /* Knows about two-pass profiling */
    }
    break;
  case SRqStartProcess: /* String Bool Bool Bool */
    {
      char cmd[10000];
#     define BA(n) GET_BOOL_VAL(getpart(p,n,4))

      p = evalArgs(p,4);
      SAVE(p,evalstring(getpart(p,0,4), cmd, sizeof cmd));
      if(debug) fprintf(stderr,"StartProcess \"%s\" ...\n",cmd);
      WHEN_NOT_REPLAY(resp=startProcess(cmd,BA(1),BA(2),BA(3)));
    }
    break;
  case SRqDLOpen: /* String */
    {
#ifdef SUPPORT_DYNLINK
      char file[10000];
      p = evalArgs(p,1);
      SAVE(p,evalstring(getpart(p,0,1), file, sizeof file));
      if(debug) fprintf(stderr,"DLOpen \"%s\" ...\n",file);
      REPLAY_ONE_AND_SKIP(p) {
	p=dlopen(file,DLOPEN_MODE);
      }
      RECORD_ONE(p);
      if(p)
	resp=mkSocketResponse(mkconstr1(SRDLHandle,mkint((int)p)));
      else
	ERR(dlerror());
#else
      ERR("DLOpen not implemented on this system");
#endif
    }
    break;
  case SRqDLClose: /* DLHandle */
    {
#ifdef SUPPORT_DYNLINK
      void *dl;
      int bool;
      p = evalArgs(p,1);
      dl = (void*)INTOF(getpart(p,0,1));
      if(debug) fprintf(stderr,"DLClose %lx ...\n",(Int)dl);
      REPLAY_ONE_AND_SKIP(bool)
	bool = !dlclose(dl);
      RECORD_ONE(bool);
      if(bool)
	resp=mkSuccess();
      else
	ERR(dlerror());
#else
      ERR("DLClose not implemented on this system");
#endif
    }
    break;
  case SRqDLSym: /* DLHandle String */
    {
#ifdef SUPPORT_DYNLINK
      char sym[10000];
      void *dl;
      void *s;
      p = evalArgs(p,2);
      dl = (void*)INTOF(getpart(p,0,2));
      SAVE(p,evalstring(getpart(p,1,2), sym, sizeof sym));
      if(debug) fprintf(stderr,"DLSym %lx \"%s\" ...",(Int)dl,sym);
      REPLAY_ONE_AND_SKIP(s)
	s = dlsym(dl,sym);
      RECORD_ONE(s);
      if(debug) fprintf(stderr," = %lx\n",(Int)s);
      if(s)
	resp=mkSocketResponse(mkconstr1(SRDLVal,mkconstr1(0,s))); /* !! */
      else
	ERR(dlerror());
#else
      ERR("DLSym not implemented on this system");
#endif
    }
    break;
  case SRqOpenFileAsSocket: /* String String */
    {
      char name[10000];
      char mode[10];
      FILE *sf;
      
      p=evalArgs(p,2);
      SAVE(p,evalstring(getpart(p,0,2),name,sizeof name));
      SAVE(p,evalstring(getpart(p,1,2),mode,sizeof mode));
      if(debug) fprintf(stderr,"OpenFileAsSocket \"%s\" \"%s\"\n",name,mode);
      REPLAY_ONE_AND_SKIP(sf)
	sf=fopen(name,mode);
      RECORD_ONE(sf);
      if(debug) fprintf(stderr,"OpenFileAsSocket returns %lx\n",(Int)sf);
      if(sf) resp=mkSocketResponse(mkSRSocket(sf));
      else ERR("OpenSocketAsFile: fopen");
    }
    break;
  case SRqWriteSocketPS: /* Socket PackedString */
    { FILE *sf;
      char *s;
      int n,len;

      p=evalArgs(p,2);
      sf = (FILE *)INTOF(getpart(p,0,2));
      getPackedStringL(getpart(p,1,2),&s,&len);
      if(debug) fprintf(stderr,"WriteSocketPS \"%.*s\"\n",len,s);
      REPLAY_ONE_AND_SKIP(n)
	n=fwrite(s,1,len,sf),fflush(sf);
      RECORD_ONE(n);
      if(debug) fprintf(stderr,"WriteSocketPS returns %d\n",n);
      resp=mkSocketResponse(mkconstr(SRWrote,1,mkint(n)));
    }
    break;
  default:
    fprintf(stderr, "Unimplemented SocketRequest %d\n", getcno(p));
    exit(1);
  }
  return resp;
}

PTR
doScall(t, p)
int t;
PTR p;
{
  PTR response;

  if(debug>2) {
    fprintf(stderr,"gc before shandle\n");
    if (t != GetAsyncInput)
      PUSHPTR(p);
    DOGC;
    if (t != GetAsyncInput)
      POPPTR(p);
  }
  switch(t) {
  case SocketRequest: response=socketrequest(p); break;
  case GetAsyncInput: response=getnextasync(); break;
  case Select: 
    asyncrequest(p);
    response=mkSuccess();
    break;
  }
  if(debug>2) {
    fprintf(stderr,"gc after shandle\n");
    GCSAVE(response);
  }
  return response;
}

void
shandle(t, p)
int t;
PTR p;
{
  mkresponse(doScall(t,p));
}
