#include <stdio.h>
#include <sys/time.h>
#include "proto.h"
#include "xmalloc.h"
#include "timeq.h"

extern void free();

struct timeq *timeq = 0;	/* time q header */
unsigned long timeqset;		/* absolute time of q header */

extern int debug;

/* Initialize timeq. (There should be a cleanup routine too) */
void
inittimeq()
{
  timeq = 0;
}

/* Make a time q entry */
struct timeq*
mktimeq(v, i, m)
int v, i, m;
{
    struct timeq *q;
    q = (struct timeq*)xmalloc(sizeof *q);
    q->value = v;
    q->interval = i;
    q->msg = m;
    q->next = 0;
    return q;
}

unsigned long
mstime()
{
    struct timeval now;
    gettimeofday(&now, (struct timezone *)0);
    return TIMETOMS(&now);
}

/* insert e in the time q */
/* value should be the offset from q start */
void
entertimeq(e)
struct timeq *e;
{
    struct timeq **p;
    unsigned long left, now;

    now = mstime();
    if (!timeq) {
	/* Time q is empty, just set it */
	timeq = e;
	e->next = 0;
    } else {
	left = now - timeqset;
	if (debug) fprintf(stderr, "Time left %ld\n", left);
	if (left >= timeq->value)
	    timeq->value = 1;
	else
	    timeq->value -= left;
	/* Find right place for it */
	for(p = &timeq; *p && (*p)->value < e->value; p = &(*p)->next)
	    e->value -= (*p)->value;
	e->next = *p;
	*p = e;
    }
    /* Restart timer */
    timeqset = now;
}

/* remove the head element in the time q, reinsert it if it has an interval */
/* return copy of the timeq element (to get msg) */
struct timeq
removetimeq()
{
    struct timeq *q;
    struct timeq r;

    if (!timeq) {
	fprintf(stderr, "removetimeq: timeq empty\n");
	exit(1);
    }
    q = timeq;
    timeq = q->next;
    r = *q;
    if (q->interval) {
	q->value = q->interval;
	entertimeq(q);
    } else {
	free(q);
    }
    return r;
}

void
cleartimeq(m)
int m;
{
    struct timeq **q;

    for(q = &timeq; *q; ) {
	if ((*q)->msg == m) {
	    struct timeq *p = *q;
	    if(debug>1) fprintf(stderr,"deleting timer %d from timeq\n",m);
	    *q = p->next;
	    free(p);
	} else {
	    q = &(*q)->next;
	}
    }
}
