#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xresource.h>
#include <X11/Xutil.h>
#include <X11/extensions/shape.h>
#include "dialog.h"
#include "heap.h"
#include "xlib.h"
#include "timeq.h"
#include "xencode.h" /* encodeEvent */
#include "xdecode.h"
#include "doxcall.h"
#include "xmalloc.h"

int LMLX = 1, _LMLX = 1;
/*
Display *display;
int screen;
*/

#define Idisplay (long)display

#define MAXPTS	1024
#define COLOR unsigned short
static GC defaultGC PROTO((Display *,Window));
static int default_bpp PROTO ((Display *,int depth));
static void getgc PROTO((PTR, GC*));
static int evalstring16 PROTO((PTR, XChar2b *, int));
static void getrect PROTO((PTR, int*, int*, int*, int*));
static void getpoint PROTO((PTR, int*, int*));
static int getpoints PROTO((PTR, XPoint []));
static void getRGB PROTO((PTR, COLOR*, COLOR*, COLOR*));
static int getButton PROTO((PTR p));
static void getGCvalues PROTO((PTR, XGCValues *, unsigned long*));
static void getWinConf PROTO((PTR, XWindowChanges *, unsigned long*));
static unsigned long geteventmask PROTO((PTR));
static XEvent getevent PROTO((Window,PTR));
static void getWinAttr PROTO((PTR, XSetWindowAttributes *, unsigned long *));
static Drawable getdrawable PROTO((Window, PTR));
static void xdodrawcommand PROTO((Display *,Window, Drawable, GC, PTR));

static PTR MkChildList  PROTO ((Window *windows,unsigned int n));
static PTR MkBitmapRead PROTO ((int,int,int,int,Pixmap));
static PTR mkVisual PROTO ((Visual *));

/* xdocommand executes an XCommand.
 * XCommands implement Xlib calls without return values.
 * Unlike other requests, XCommands DO NOT have responses.
 */
static void xdocommand PROTO((Display *, Window, PTR));

/* xdorequest executes an XRequest.
 * XRequests implement Xlib calls with return values.
 * xdorequest(...) constructs & returns an appropriate XResponse.
 */
static PTR xdorequest PROTO((Display *, Window, PTR)); /* ?? */


#define DEFAULT_EVENT_MASK ((1<<25)-1-PointerMotionHintMask)	/* This should cover all events !! */

#define DEFAULT_FONT  "-*-courier-bold-r-normal--14-*-*-*-*-*-*-*"

#define DEFAULT_GRAB_MASK (ButtonPressMask|ButtonReleaseMask|ButtonMotionMask)


static struct {
  Display *display;
  Window window;
  Pixmap pixmap[2];
} gcwarning;

/* static */
void gcwarninghook(n)
     int n;
{
  if(gcwarning.display) {
    XSetWindowBackgroundPixmap(gcwarning.display,
			       gcwarning.window,
			       gcwarning.pixmap[n!=0]);
    XClearWindow(gcwarning.display,gcwarning.window);
    XFlush(gcwarning.display);
  }
}

static GC defaultGC(display,w)  /* Never called in two-pass profiling */
     Display *display;
     Window w;
{
  GC gc;
  int screen;
  screen=DefaultScreen(display); /* which screen?! */
  gc=XCreateGC(display,w,0,NULL);
  XSetForeground(display,gc,BlackPixel(display,screen));
  XSetBackground(display,gc,WhitePixel(display,screen));
  return gc;
}

static int default_bpp(display,depth)
     Display *display;
     int depth;
{
  int i,cnt,bpp;
  XPixmapFormatValues *ps=XListPixmapFormats(display,&cnt);
  bpp=depth; /* Hmm. Something is wrong if depth isn't found. */
  for(i=0;i<cnt;i++)
    if(ps[i].depth==depth) {
      bpp=ps[i].bits_per_pixel;
      break;
    }
  XFree(ps);
  if(debug>1) fprintf(stderr,"default_bpp(display,%d)=%d\n",depth,bpp);
  return bpp;
}

static void
getgc(gp, gcp)
PTR gp;
GC *gcp;
{
    *gcp = (GC)INTOF(gp);
}

PTR
evalArgs(p,n)
PTR p;
int n;
{
  int c;
  for(c=0;c<n;c++) {
    PUSHPTR(p);
    (void)evaluate(getpart(p,c,n));
    POPPTR(p);
  }
  return p;
}

static int
evalstring16(s,buf,n)
PTR s;
XChar2b *buf;
int n; /* buffer size */
{
  int c;
  XChar2b *p=buf;
 
  s = evaluate(s);
  while(!ISNIL(s) && n-->0) {
    SAVE(s,c=CHAROF(EHDOF(s)));
    p->byte1=c>>8;
    p->byte2=c;
    p++;
    s = ETLOF(s);
  }
  return p-buf;
}

static void
getrect(rp, x1, y1, x2, y2)
PTR rp;
int *x1, *y1, *x2, *y2;
{
  rp = evalArgs(rp,2);
  SAVE(rp,getpoint(FSTOF(rp),x1,y1));
  getpoint(SNDOF(rp),x2,y2);
}

static XRectangle
getxrect(rp)
PTR rp;
{
  int x1,y1,x2,y2;
  XRectangle xr;
  getrect(rp,&x1,&y1,&x2,&y2);
  xr.x=x1; 
  xr.y=y1; 
  xr.width=x2; 
  xr.height=y2;
  return xr;
}

static void
getpoint(pp, xp, yp)
PTR pp;
int *xp, *yp;
{
    pp = evalArgs(pp,2);
    *xp = INTOF(FSTOF(pp));
    *yp = INTOF(SNDOF(pp));
}

static void
getRGB(pp, r,g,b)
PTR pp;
COLOR *r,*g,*b;
{
  pp = evalArgs(pp,3);
    *r = INTOF(getpart(pp,0,3));
    *g = INTOF(getpart(pp,1,3));
    *b = INTOF(getpart(pp,2,3));
}

static int
getButton(q)
PTR q;
{
  if (getcno(q) == 0)
    return AnyButton;
  else
    return INTOF(EARG1(q));
}

static void
getGCvalues(p, vp, vmp)
PTR p;
XGCValues *vp;
unsigned long *vmp;
{
  PTR f;
  
  *vmp = 0;
  for(;!ISNIL(p);p = ETLOF(p)) {
    PUSHPTR(p);
    f = EHDOF(p);
    switch(getcno(f)) {
    case 0:			/* GCFunction */
      vp->function = GET_ENUM(EARG1(f));
      *vmp |= GCFunction;
      break;
    case 1:			/* GCForeGround */
      vp->foreground = INTOF(EARG1(f));
      *vmp |= GCForeground;
      break;
    case 2:			/* GCBackGround */
      vp->background = INTOF(EARG1(f));
      *vmp |= GCBackground;
      break;
    case 3:			/* GCLineWidth */
      vp->line_width = INTOF(EARG1(f));
      *vmp |= GCLineWidth;
      break;
    case 4:			/* GCLineStyle */
      vp->line_style = GET_ENUM(EARG1(f));
      *vmp |= GCLineStyle;
      break;
    case 5:			/* GCFont */
      vp->font = INTOF(EARG1(f));
      *vmp |= GCFont;
      break;
    case 6:			/* GCCapStyle */
      vp->cap_style = GET_ENUM(EARG1(f));
      *vmp |= GCCapStyle;
      break;
    case 7:                 /* GCSubwindowMode */
      vp->subwindow_mode = GET_ENUM(EARG1(f));
      *vmp |= GCSubwindowMode;
      break;
    case 8:               /* GCGraphicsExposures Bool */
      vp->graphics_exposures = GET_BOOL_VAL(EARG1(f));
      *vmp |= GCGraphicsExposures;
      break;
    case 9:               /* GCFillStyle GCFillStyle */
      vp->fill_style = GET_ENUM(EARG1(f));
      *vmp |= GCFillStyle;
      break;
    case 10:              /* GCTile PixmapId */
      vp->tile = INTOF(EARG1(f));
      *vmp |= GCTile;
      break;
    case 11:		  /* GCStipple PixmapId */
      vp->stipple = INTOF(EARG1(f));
      *vmp |= GCStipple;
      break;
    default:
      fprintf(stderr, "Bad X GC command %d\n", getcno(f));
      exit(1);
    }
    POPPTR(p);
  }
}

static void
getWinConf(p, vp, vmp)
PTR p;
XWindowChanges *vp;
unsigned long *vmp;
{
    PTR f;

    *vmp = 0;
    for(;!ISNIL(p);p = ETLOF(p)) {
        PUSHPTR(p);
	f = EHDOF(p);
	switch(getcno(f)) {
	case 0:			/* CWX */
	    vp->x = INTOF(EARG1(f));
	    *vmp |= CWX;
	    break;
	case 1:			/* CWY */
	    vp->y = INTOF(EARG1(f));
	    *vmp |= CWY;
	    break;
	case 2:			/* CWWidth */
	    vp->width = INTOF(EARG1(f));
	    *vmp |= CWWidth;
	    break;
	case 3:			/* CWHeight */
	    vp->height = INTOF(EARG1(f));
	    *vmp |= CWHeight;
	    break;
	case 4:			/* CWBorderWidth */
	    vp->border_width = INTOF(EARG1(f));
	    *vmp |= CWBorderWidth;
	    break;
	case 5:			/* CWStackMode */
	    vp->stack_mode = GET_ENUM(EARG1(f));
	    *vmp |= CWStackMode;
	    break;
	default:
	    fprintf(stderr, "Bad X window config command %d\n", getcno(f));
	    exit(1);
	}
	POPPTR(p);
    }
}

static unsigned long
geteventmask(p)
PTR p;
{
    PTR f;
    unsigned long m = 0;

    for(;!ISNIL(p);p = ETLOF(p)) {
        PUSHPTR(p);
	f = EHDOF(p);
	m |= 1<<GET_ENUM(f);
	POPPTR(p);
    }
    if(debug) fprintf(stderr, "emask %lx\n", m);
    return m;
}

static XEvent
getevent(w,p)
Window w;
PTR p;
{
  XEvent e;
  e.xany.window=w;
  /* Only implemented for SelectionNotify... */
  switch(getcno(p)) {
  case XESelectionNotify:
    {
      PTR f = evalArgs(getpart(p,1,2),3);
      e.type = SelectionNotify;
      e.xselection.selection = ATOMOF(getpart(f,0,3));
      e.xselection.target = ATOMOF(getpart(f,1,3));
      e.xselection.property = ATOMOF(getpart(f,2,3));
      e.xselection.time = INTOF(getpart(p,0,2));
    }
    break;
  default:
    fprintf(stderr,"getevent on %d (only implemented for SelectionNotify)\n",e.type);
    exit(1);
  }
  return e;
}

static void
getWinAttr(p, ap, vmp)
PTR p;
XSetWindowAttributes *ap;
unsigned long *vmp;
{
    PTR f;

    *vmp = 0;
    for(;!ISNIL(p);p = ETLOF(p)) {
        PUSHPTR(p);
	f = EHDOF(p);
	switch(getcno(f)) {
	case 0:			/* CWEventMask */
	    ap->event_mask = geteventmask(EARG1(f));
	    *vmp |= CWEventMask;
	    break;
	case 1:			/* CWBackingStore */
	    ap->backing_store = GET_ENUM(EARG1(f));
	    *vmp |= CWBackingStore;
	    break;
	case 2:			/* CWSaveUnder */
	    ap->save_under = GET_BOOL_VAL(EARG1(f));
	    *vmp |= CWSaveUnder;
	    break;
	case 3:			/* CWDontPropagate */
	    ap->do_not_propagate_mask = geteventmask(EARG1(f));
	    *vmp |= CWDontPropagate;
	    break;
	case 4:			/* CWOverrideRedirect */
	    ap->override_redirect = GET_BOOL_VAL(EARG1(f));
	    *vmp |= CWOverrideRedirect;
	    break;
	case 5:			/* CWBackPixel */
	    ap->background_pixel = INTOF(EARG1(f));
	    *vmp |= CWBackPixel;
	    break;
	case 6:			/* CWCursor */
	    ap->cursor = INTOF(EARG1(f));
	    *vmp |= CWCursor;
	    break;
	case 7:			/* CWBitGravity */
	    ap->bit_gravity = GET_ENUM(EARG1(f));
	    *vmp |= CWBitGravity;
	    break;
	case 8:			/* CWWinGravity */
	    ap->win_gravity = GET_ENUM(EARG1(f));
	    *vmp |= CWWinGravity;
	    break;
	case 9:			/* CWBackPixmap */
	    if(debug) fprintf(stderr,"CWBackPixmap\n");
	    ap->background_pixmap = INTOF(EARG1(f));
	    *vmp |= CWBackPixmap;
	    break;
	case 10:			/* CWBorderPixmap */
	    if(debug) fprintf(stderr,"CWBorderPixmap\n");
	    ap->border_pixmap = INTOF(EARG1(f));
	    *vmp |= CWBorderPixmap;
	    break;
	case 11:			/* CWBorderPixel */
	    if(debug) fprintf(stderr,"CWBorderPixel\n");
	    ap->border_pixel = INTOF(EARG1(f));
	    *vmp |= CWBorderPixel;
	    break;
	default:
	    fprintf(stderr, "Bad X window attr %d\n", getcno(f));
	    exit(1);
	}
	POPPTR(p);
    }
}

Drawable getdrawable(wi,p)
Window wi;
PTR p;
{
  /* This could actually be handled in the LML program... */

  if (debug > 1) fprintf(stderr, "getdrawable %d\n", getcno(p));
  switch(getcno(p)) {
    case XDrMyWindow:
      return wi;
    case XDrPixmap:
      return INTOF(EARG1(p));
    default:
      /*if (debug)*/ fprintf(stderr, "Bad Drawable %d\n", getcno(p));
      exit(1);
  }
}

static int
getpoints(l,points)
PTR l;
XPoint points[];
{
  int n;
  int x,y;

  for(n=0;!ISNIL(l); l = ETLOF(l),n++) {
    SAVE(l,getpoint(EHDOF(l), &x, &y));
    points[n].x = x; points[n].y = y;
  }
  return n;
}

void
xdodrawcommand(display,realwi,wi,gc,p)
Display *display;
Window realwi;
Drawable wi;
GC gc;
PTR p;
{
  if (debug > 1) fprintf(stderr, "xdodrawcommand %d\n", getcno(p));
  switch(getcno(p)) {
  case XDCDrawLine:		/* Line */
    {
      int x1,y1,x2,y2;
      
      getrect(EARG1(p), &x1, &y1, &x2, &y2);
      if (debug) fprintf(stderr, "XDrawLine(%lx,%lx,%lx,%d,%d,%d,%d)\n", Idisplay, wi,(long)gc,x1,y1,x2,y2);
      WHEN_NOT_REPLAY(XDrawLine(display,wi,gc,x1,y1,x2,y2));
    }
    break;
  case XDCDrawLines:		/* CoordMode [Point] */
    {
      int mode;
      XPoint point[MAXPTS];
      int n;
      
      p = evalArgs(p,2);
      mode = GET_ENUM(getpart(p, 0, 2));
      n = getpoints(getpart(p,1,2),point);

      if (debug) fprintf(stderr, "XDrawLines(%lx,%lx,%lx,...,%d,%d)\n",Idisplay,wi,(long)gc,n,mode);
      WHEN_NOT_REPLAY(XDrawLines(display,wi,gc,point,n,mode));
    }
    break;
  case XDCDrawImageString:	/* Point String */
    {
      int x,y;
      char buf[10000];

      p = evalArgs(p,2);
      SAVE(p,getpoint(getpart(p, 0, 2), &x, &y));
      evalstring(getpart(p, 1, 2), buf, sizeof buf);
      if (debug) fprintf(stderr, "XDrawImageString(%lx,%lx,%lx,%d,%d,\"%s\")\n", Idisplay, wi,(long)gc,x,y,buf);
      WHEN_NOT_REPLAY(XDrawImageString(display,wi,gc,x,y,buf,strlen(buf)));
    }
    break;
  case XDCDrawImageString16:	/* Point String */
    {
      int x,y;
      XChar2b buf[5000];
      int length;
      
      p = evalArgs(p,2);
      SAVE(p,getpoint(getpart(p, 0, 2), &x, &y));
      length=evalstring16(getpart(p, 1, 2), buf, sizeof buf/sizeof(XChar2b));
      if (debug) fprintf(stderr, "XDrawImageString16(%lx,%lx,%lx,%d,%d,...)\n", Idisplay, wi,(long)gc,x,y/*,buf*/);
      WHEN_NOT_REPLAY(XDrawImageString16(display,wi,gc,x,y,buf,length));
    }
    break;
  case XDCDrawImageStringPS:	/* Point PackedString */
    {
      int x,y;
      char *s;
      unsigned len;

      p = evalArgs(p,2);
      SAVE(p,getpoint(getpart(p, 0, 2), &x, &y));
      getPackedStringL(getpart(p,1,2),&s,&len);
      if (debug) fprintf(stderr, "PS XDrawImageString(%lx,%lx,%lx,%d,%d,\"%*s\")\n", Idisplay, wi,(long)gc,x,y,len,s);
      WHEN_NOT_REPLAY(XDrawImageString(display,wi,gc,x,y,s,len));
    }
    break;
  case XDCDrawString:	/* Point String */
    {
      int x,y;
      char buf[10000];
      
      p = evalArgs(p,2);
      SAVE(p,getpoint(getpart(p, 0, 2), &x, &y));
      evalstring(getpart(p, 1, 2), buf, sizeof buf);
      if (debug) fprintf(stderr, "XDrawString(%lx,%lx,%lx,%d,%d,\"%s\")\n", Idisplay, wi,(long)gc,x,y,buf);
      WHEN_NOT_REPLAY(XDrawString(display,wi,gc,x,y,buf,strlen(buf)));
    }
    break;
  case XDCDrawString16:	/* Point String */
    {
      int x,y;
      XChar2b buf[5000];
      int length;
      
      p = evalArgs(p,2);
      SAVE(p,getpoint(getpart(p, 0, 2), &x, &y));
      length=evalstring16(getpart(p, 1, 2), buf, sizeof buf/sizeof(XChar2b));
      if (debug) fprintf(stderr, "XDrawString16(%lx,%lx,%lx,%d,%d,...)\n", Idisplay, wi,(long)gc,x,y/*,buf*/);
      WHEN_NOT_REPLAY(XDrawString16(display,wi,gc,x,y,buf,length));
    }
    break;
  case XDCDrawStringPS:	/* Point PackedString */
    {
      int x,y;
      char *s;
      unsigned len;
      
      p = evalArgs(p,2);
      SAVE(p,getpoint(getpart(p, 0, 2), &x, &y));
      getPackedStringL(getpart(p,1,2),&s,&len);
      if (debug) fprintf(stderr, "PS XDrawString(%lx,%lx,%lx,%d,%d,\"%*s\")\n", Idisplay, wi,(long)gc,x,y,len,s);
      WHEN_NOT_REPLAY(XDrawString(display,wi,gc,x,y,s,len));
    }
    break;
  case XDCDrawRectangle:	/* Rect */
    {
      int x1,y1,x2,y2;
      
      getrect(EARG1(p), &x1, &y1, &x2, &y2);
      if (debug) fprintf(stderr, "XDrawRectangle(%lx,%lx,%lx,%d,%d,%d,%d)\n", Idisplay, wi,(long)gc,x1,y1,x2,y2);
      WHEN_NOT_REPLAY(XDrawRectangle(display,wi,gc,x1,y1,x2,y2));
    }
    break;
  case XDCFillRectangle:	/* Rect */
    {
      int x1,y1,x2,y2;
      
      getrect(EARG1(p), &x1, &y1, &x2, &y2);
      if (debug) fprintf(stderr, "XFillRectangle(%lx,%lx,%lx,%d,%d,%d,%d)\n", Idisplay, wi,(long)gc,x1,y1,x2,y2);
      WHEN_NOT_REPLAY(XFillRectangle(display,wi,gc,x1,y1,x2,y2));
    }
    break;
  case XDCFillPolygon:		/* Shape CoordMode [Point] */
    {
      int mode, shape;
      XPoint point[MAXPTS];
      int n;
      
      p = evalArgs(p,3);
      shape = GET_ENUM(getpart(p, 0, 3));
      mode = GET_ENUM(getpart(p, 1, 3));
      n = getpoints(getpart(p,2,3),point);

      if (debug) fprintf(stderr, "XFillPolygon(%lx,%lx,%lx,...,%d,%d,%d)\n",Idisplay,wi,(long)gc,n,shape,mode);
      WHEN_NOT_REPLAY(XFillPolygon(display,wi,gc,point,n,shape,mode));
    }
    break;
  case XDCDrawArc:		/* Rect Int Int */
    {
      int x,y,w,h,angle1,angle2;
      
      p = evalArgs(p,3);
      SAVE(p,getrect(getpart(p, 0, 3), &x, &y, &w, &h));
      angle1=INTOF(getpart(p, 1, 3));
      angle2=INTOF(getpart(p, 2, 3));
      if (debug) fprintf(stderr, "XDrawArc(%lx,%lx,%lx,%d,%d,%d,%d,%d,%d)\n", Idisplay, wi,(long)gc,x,y,w,h,angle1,angle2);
      WHEN_NOT_REPLAY(XDrawArc(display,wi,gc,x,y,w,h,angle1,angle2));
    }
    break;
  case XDCFillArc:		/* Rect Int Int */
    {
      int x,y,w,h,angle1,angle2;
      
      p = evalArgs(p,3);
      SAVE(p,getrect(getpart(p, 0, 3), &x, &y, &w, &h));
      angle1=INTOF(getpart(p, 1, 3));
      angle2=INTOF(getpart(p, 2, 3));
      if (debug) fprintf(stderr, "XFillArc(%lx,%lx,%lx,%d,%d,%d,%d,%d,%d)\n", Idisplay, wi,(long)gc,x,y,w,h,angle1,angle2);
      WHEN_NOT_REPLAY(XFillArc(display,wi,gc,x,y,w,h,angle1,angle2));
    }
    break;
  case XDCCopyArea:	/* src Rect Point */
    {
      Drawable src;
      int src_x,src_y,w,h,dst_x,dst_y;
      
      p = evalArgs(p,3);
      SAVE(p,src=getdrawable(realwi,getpart(p,0,3)));
      SAVE(p,getrect(getpart(p,1,3),&src_x,&src_y,&w,&h));
      getpoint(getpart(p,2,3),&dst_x,&dst_y);
      if (debug) fprintf(stderr, "XCopyArea(%lx,%lx,%lx,%lx,%d,%d,%d,%d,%d,%d)\n", Idisplay, src, wi,(long)gc,src_x,src_y,w,h,dst_x,dst_y);
      WHEN_NOT_REPLAY(XCopyArea(display,src,wi,gc,src_x,src_y,w,h,dst_x,dst_y));
    }
    break;
  case XDCCopyPlane:	/* src Rect Point plane */
    {
      Drawable src;
      int src_x,src_y,w,h,dst_x,dst_y;
      unsigned long plane;
      
      p = evalArgs(p,3);
      SAVE(p,src=getdrawable(realwi,getpart(p,0,4)));
      SAVE(p,getrect(getpart(p,1,4),&src_x,&src_y,&w,&h));
      getpoint(getpart(p,2,4),&dst_x,&dst_y);
      plane = 1 << INTOF(getpart(p,3,4));
      if (debug) fprintf(stderr, "XCopyPlane(%lx,%lx,%lx,%lx,%d,%d,%d,%d,%d,%d,%lx)\n", Idisplay, src, wi,(long)gc,src_x,src_y,w,h,dst_x,dst_y,plane);
      WHEN_NOT_REPLAY(XCopyPlane(display,src,wi,gc,src_x,src_y,w,h,dst_x,dst_y,plane));
    }
    break;
  case XDCDrawPoint:
    {
      int x,y;
      
      getpoint(EARG1(p), &x, &y);
      if (debug) fprintf(stderr, "XDrawPoint(%lx,%lx,%lx,%d,%d)\n", Idisplay, wi,(long)gc,x,y);
      WHEN_NOT_REPLAY(XDrawPoint(display,wi,gc,x,y));
    }
    break;
  case XDCCreatePutImage: /* Rect ImageFormat [Pixel] */
    {
      int x,y,w,h;
      XImage *image;
      int screen = DefaultScreen(display);
      int format;
      int depth = DefaultDepth(display, screen);
      int bpp = default_bpp(display,depth);
      int Bdepth = (depth+7)>>3; /* hmm. bytes per pixel */
      int Bpp = (bpp+7)>>3;
      int bitmap_pad = 32;
      PTR list;
      int size;
      char *data;
      int nullCount;
      int bytes_per_line;
  
      p = evalArgs(p,3);
      SAVE(p,getrect(getpart(p, 0, 3), &x, &y,&w,&h));
      bytes_per_line = (w*Bpp + 3) & ~3; /* assumes bitmap_pad==32 */
      nullCount = bytes_per_line - w*Bpp;
      /*fprintf(stderr,"CreatePutImage: nullCount=%d\n",nullCount);*/
      format = INTOF(getpart(p,1,3));
      list = getpart(p,2,3);
      size = w*h;
      data = xmalloc(bytes_per_line*h);
      {
	int i,j,k;
	unsigned b;
	char *ip;
	for (i=0, ip=data; i<h; i++) {
	  for (j=0; j<w; j++) {
	    SAVE(list,b = ISNIL(list) ? 0 : INTOF(EHDOF(list)));
	    if(!ISNIL(list)) list = ETLOF(list);
	    if(ImageByteOrder(display)==LSBFirst) {
	      for(k=0;k<Bdepth;k++) {
		*ip++=b; b>>=8; /* LSBFirst */
	      }
	      ip+=Bpp-Bdepth;
	    }
	    else {
	      ip+=Bpp-Bdepth;
	      for(k=Bdepth-1;k>=0;k--) {
		ip[k]=b; b>>=8; /* MSBFirst */
	      }
	      ip+=Bdepth;
	    }
	  }
	  for (j=0; j<nullCount; j++)  *ip++ = 0;
	}
      }
    WHEN_NOT_REPLAY_START
      image = XCreateImage(display,DefaultVisual(display,screen),depth,format,0,data,w,h,bitmap_pad,bytes_per_line);
      if (debug) fprintf(stderr, "XCreateImage(format=%d,w=%d,h=%d) = %lx\n", format,w,h,(Int)image);
      if(image) {
	XPutImage(display,wi,gc,image,0,0,x,y,w,h);
	if (debug) fprintf(stderr, "XPutImage(gc=%lx,x=%d,y=%d)\n",(Int)gc,x,y);
	XDestroyImage(image);
      }
    WHEN_NOT_REPLAY_END
      /*free(data);*/
     /*XDrestroyImage() calls free() for images created with XCreateImage().*/
    }
    break;
  default:
    /*if (debug)*/ fprintf(stderr, "Unimplemented Draw Command %d\n", getcno(p));
  }
}

void
xdocommand(display, wi, p)
     Display *display;
     Window wi;
     PTR p;
{
  if (debug > 1) fprintf(stderr, "xdocommand %d\n", getcno(p));
  switch(getcno(p)) {
  case XCCloseDisplay: /* Display */
    { 
      Display *disp = (Display *)INTOF(EARG1(p));
      if (debug) fprintf(stderr, "XCloseDisplay(%lx)\n", (Int)disp);
      WHEN_NOT_REPLAY(XCloseDisplay(disp));
    }
    break;
  case XCDestroyWindow:
    if (debug) fprintf(stderr, "XDestroyWindow(%lx,%lx)\n", Idisplay, wi);
    WHEN_NOT_REPLAY(XDestroyWindow(display,wi));
    break;
  case XCMapRaised:
    if (debug) fprintf(stderr, "XMapRaised(%lx,%lx)\n", Idisplay, (long)wi);
    WHEN_NOT_REPLAY(XMapRaised(display,wi));
    break;
  case XCLowerWindow:
    if (debug) fprintf(stderr, "XLowerWindow(%lx,%lx)\n", Idisplay, (long)wi);
    WHEN_NOT_REPLAY(XLowerWindow(display,wi));
    break;
  case XCUnmapWindow:
    if (debug) fprintf(stderr, "XUnmapWindow(%lx,%lx)\n", Idisplay, wi);
    WHEN_NOT_REPLAY(XUnmapWindow(display,wi));
    break;
  case XCDraw:	/* Drawable GCId DrawCommand */
      {
	Drawable dr;
	GC gc;
	p = evalArgs(p,3);
	SAVE(p,dr = getdrawable(wi,getpart(p,0,3)));
	gc = (GC)INTOF(getpart(p,1,3));
	(void)xdodrawcommand(display, wi, dr,gc,getpart(p,2,3));
      }
    break;
  case XCClearArea:		/* tRect Bool */
    {
      int x,y,width,height,exposures;
     
      p = evalArgs(p,2);
      SAVE(p,getrect(ARG2(p,1), &x,&y,&width,&height));
      exposures = GET_BOOL_VAL(ARG2(p,2));
      if (debug) fprintf(stderr, "XClearArea(%lx,%lx,%d,%d,%d,%d,%d)\n", Idisplay, wi,x,y,width,height,exposures);
      WHEN_NOT_REPLAY(XClearArea(display,wi,x,y,width,height,exposures));
    }
    break;
  case XCClearWindow:
    if (debug) fprintf(stderr, "XClearWindow(%lx,%lx)\n", Idisplay, wi);
    WHEN_NOT_REPLAY(XClearWindow(display,wi));
    break;
#if 0
  case XCChangeGC:		/* tGCId tGCAttributeList  */
    break;
#endif
  case XCFreeGC:		/* tGCId */
    {
      GC gc;
      
      getgc(EARG1(p), &gc);
      if (debug) fprintf(stderr, "XFreeGC(%lx,%lx)\n", Idisplay, (long)gc);
      WHEN_NOT_REPLAY(XFreeGC(display,gc));
    }
    break;
  case XCChangeWindowAttributes: /* List tWindowAttributes */
    {
      XSetWindowAttributes attr;
      unsigned long valuemask;
      
      getWinAttr(EARG1(p), &attr, &valuemask);
      if (debug) fprintf(stderr, "XCChangeWindowAttributes(%lx,%lx,%lx,...)\n",Idisplay,wi,valuemask);
      WHEN_NOT_REPLAY(XChangeWindowAttributes(display,wi,valuemask,&attr));
    }
    break;
  case XCConfigureWindow:	/* List tWindowChanges */
    {
      XWindowChanges changes;
      unsigned long valuemask;
      
      getWinConf(EARG1(p), &changes, &valuemask);
      if (debug) fprintf(stderr, "XCConfigureWindow(%lx,%lx,%lx,...)\n", Idisplay,wi,valuemask);
      WHEN_NOT_REPLAY(XConfigureWindow(display,wi,valuemask,&changes));
    }
    break;
  case XCStoreName:
    {
      char buf[1000];
      evalstring(ARG1(p), buf, sizeof buf);
      if (debug) fprintf(stderr, "XStoreName(%lx,%lx,%s)\n", Idisplay,wi,buf);
      WHEN_NOT_REPLAY(XStoreName(display,wi,buf));
    }
    break;
  case XCSetNormalHints:	/* tPoint */
    {
      XSizeHints h;
      int x, y;
      
      getpoint(EARG1(p), &x, &y);
      h.x = x; h.y = y;
      if (debug) fprintf(stderr, "XSetNormalHints(%lx,%lx,...)\n",Idisplay,wi);
      h.flags = USPosition;
      WHEN_NOT_REPLAY(XSetNormalHints(display,wi,&h));
    }
    break;
  case XCSetWMHints:	/* Bool */
    {
      XWMHints h;
      
      h.flags = InputHint;
      h.input = GET_BOOL_VAL(EARG1(p));
      if (debug) fprintf(stderr, "XSetWMHints(%lx,%lx,%d)\n",Idisplay,wi,h.input);
      WHEN_NOT_REPLAY(XSetWMHints(display,wi,&h));
    }
    break;
  case XCUngrabPointer:
    if (debug) fprintf(stderr, "XUngrabPointer(%lx,CurrentTime)\n", Idisplay);
    WHEN_NOT_REPLAY(XUngrabPointer(display,CurrentTime));
    break;
  case XCGrabButton:		/* Bool Button ModState [EventMask] */
    {
      Window confine_to=None;
      Bool owner_events;
      unsigned int button,
      modifiers=AnyModifier, /* not implemented! */
      event_mask;
      int pointer_mode=GrabModeAsync,keyboard_mode=GrabModeAsync;
      Cursor cursor=None;
      
      p = evalArgs(p,4);
      owner_events = GET_BOOL_VAL(ARG4(p, 1));
      SAVE(p,button = getButton(ARG4(p, 2)));
      SAVE(p,event_mask = geteventmask(ARG4(p, 4)));
      if (debug) fprintf(stderr, "XGrabButton(%lx,%d,%x,%lx,%d,%x,%d,%d,%lx,%lx)\n",Idisplay,button,modifiers,wi,owner_events,event_mask,
			 pointer_mode,keyboard_mode,confine_to,cursor);
      WHEN_NOT_REPLAY(XGrabButton(display,button,modifiers,wi,owner_events,event_mask,pointer_mode,keyboard_mode,confine_to,cursor));
    }
    break;
  case XCUngrabButton:        /* Button ModState */
    {
      unsigned int button,modifiers;
      p = evalArgs(p,2);
      SAVE(p,button=getButton(ARG2(p,1)));
      modifiers=AnyModifier; /* !!! */
      if (debug) fprintf(stderr, "XUngrabButton(%lx,%d,%x,%lx)\n",Idisplay,button,modifiers,wi);
      WHEN_NOT_REPLAY(XUngrabButton(display,button,modifiers,wi));
    }
    break;
  case XCFlush:
    if (debug) fprintf(stderr, "XFlush(%lx)\n", Idisplay);
    WHEN_NOT_REPLAY(XFlush(display));
    break;
  case XCFreePixmap:		/* Pixmap */
    {
      Pixmap pixmap=INTOF(EARG1(p));
      if(debug) fprintf(stderr,"XFreePixmap(%lx,%lx)\n",Idisplay,pixmap);
      WHEN_NOT_REPLAY(XFreePixmap(display,pixmap));
    }
    break;
  case XCShapeCombineMask:	/* ShapeKind Point PixmapId ShapeOperation */
    {
      int destKind,x,y,op;
      Pixmap pixmap;

      p = evalArgs(p,4);
      destKind=GET_ENUM(getpart(p,0,4));
      pixmap=INTOF(getpart(p,2,4));
      op=GET_ENUM(getpart(p,3,4));
      getpoint(getpart(p,1,4),&x,&y);
      if(debug) fprintf(stderr,"XShapeCombineMask(%lx,%lx,%x,%d,%d,%lx,%d)\n",Idisplay,wi,destKind,x,y,pixmap,op);
      WHEN_NOT_REPLAY(XShapeCombineMask(display,wi,destKind,x,y,pixmap,op));
    }
    break;
  case XCShapeCombineRectangles: 
       /* ShapeKind Point [Rect] ShapeOperation Ordering */
    {
      int destKind,x,y,op,ord;
      int nrects;
      XRectangle rects[MAXPTS];

      p = evalArgs(p,5);
      destKind=GET_ENUM(getpart(p,0,5));
      getpoint(getpart(p,1,5),&x,&y);
      {
	PTR l;
	
	l = getpart(p, 2, 5);
	nrects = 0;
	for(;!ISNIL(l); l = ETLOF(l)) {
	    /*int x,y,w,h;*/
	  SAVE(l,rects[nrects]=getxrect(EHDOF(l)));
	  nrects++;
	}
      }
      op=GET_ENUM(getpart(p,3,5));
      ord=GET_ENUM(getpart(p,4,5));
      if(debug) fprintf(stderr,"XShapeCombineRectangles(%lx,%lx,%x,%d,%d,...,%d,%d,%d)\n",Idisplay,wi,destKind,x,y,nrects,op,ord);
      WHEN_NOT_REPLAY(XShapeCombineRectangles(display,wi,destKind,x,y,rects,nrects,op,ord));
    }
    break;
  case XCShapeCombineShape: 
       /* ShapeKind Point PixmapId ShapeKind ShapeOperation */
    {
      int destKind,srcKind,x,y,op;
      Pixmap pixmap;

      p = evalArgs(p,5);
      destKind=GET_ENUM(getpart(p,0,5));
      getpoint(getpart(p,1,5),&x,&y);
      pixmap=INTOF(getpart(p,2,5));
      srcKind=GET_ENUM(getpart(p,3,5));
      op=GET_ENUM(getpart(p,4,5));
      if(debug) fprintf(stderr,"XShapeCombineShape(%lx,%lx,%x,%d,%d,%lx,%d,%d)\n",Idisplay,wi,destKind,x,y,pixmap,srcKind,op);
      WHEN_NOT_REPLAY(XShapeCombineShape(display,wi,destKind,x,y,pixmap,srcKind,op));
    }
    break;
  case XCRmDestroyDatabase: /* RmDatabase */
    {
      XrmDatabase db = (XrmDatabase)INTOF(EARG1(p));
      if (debug) fprintf(stderr, "XrmDestroyDatabase(\"%lx\")\n",(long)db);
      WHEN_NOT_REPLAY(XrmDestroyDatabase(db));
    }
    break;
#if 0
    /* Doesn't work in X11R4. */
  case XCRmCombineDatabase: /* RmDatabase RmDatabase Bool */
    {
      XrmDatabase source,target;
      Bool override;
      
      p = evalArgs(p,3);
      source = (XrmDatabase)INTOF(getpart(p,0,3));
      target = (XrmDatabase)INTOF(getpart(p,1,3));
      override = GET_BOOL_VAL(getpart(p,2,3));
      if (debug) fprintf(stderr, "XrmCombineDatabase(\"%lx,%lx,%ld\")\n",(long)source,(long)target,(long)override);
      WHEN_NOT_REPLAY(XrmCombineDatabase(source,&target,override));
    }
    break;
#endif
  case XCRmPutLineResource: /* RmDatabase String */
    {
      char buf[1000];
      XrmDatabase db;

      p = evalArgs(p,2);
      db = (XrmDatabase)INTOF(ARG2(p,1));
      evalstring(ARG2(p,2), buf, sizeof buf);
      if (debug) fprintf(stderr, "XrmPutLineResource(%lx,\"%s\")\n",(long)db, buf);
      WHEN_NOT_REPLAY(XrmPutLineResource(&db, buf)); /* Sorry, only existing db's tolerated */
    }
    break;
  case XCSetWMProtocols: /* [Atom] */
    {
      Atom protocols[1000];
      PTR l;
      int n;
      
      for(n=0,l=EARG1(p);!ISNIL(l); n++,l=ETLOF(l))
	SAVE(l,protocols[n]=ATOMOF(EHDOF(l)));
      
      if (debug) fprintf(stderr, "XSetWMProtocols(%lx,%lx,...,%d)\n",Idisplay, wi, n);
      WHEN_NOT_REPLAY(XSetWMProtocols(display,wi,protocols,n)); /* returns Status... */
    }
    break;
  case XCSendEvent: /* Window Bool [EventMask] Event */
    {
      Window w;
      Bool propagate;
      long eventmask;
      XEvent event;

      p = evalArgs(p,4);
      w = INTOF(getpart(p,0,4));
      propagate = GET_BOOL_VAL(getpart(p,1,4));
      SAVE(p,eventmask = geteventmask(getpart(p,2,4)));
      event = getevent(w,getpart(p,3,4));
      if (debug) fprintf(stderr, "XSendEvent(%lx,%lx,%d,%lx)\n",Idisplay,w,propagate,eventmask);
      {
	int s;
	REPLAY_ONE_AND_SKIP(s)
	  s = XSendEvent(display,w,propagate,eventmask,&event);
	RECORD_ONE(s);
	if(!s) fprintf(stderr,"XSendEvent returned %d\n",s);
      }
    }
    break;
  case XCSetSelectionOwner: /* Bool Atom */
    {
      Bool getit;
      Window w;
      Atom selection;
      Time time = CurrentTime;

      p = evalArgs(p,2);
      getit = GET_BOOL_VAL(getpart(p,0,2));
      w = getit ? wi : None;
      selection = ATOMOF(getpart(p,1,2));
      if (debug) fprintf(stderr, "XSetSelectionOwner(%lx,%lx,%lx)\n",(Int)display,(Int)selection,(Int)w);
      WHEN_NOT_REPLAY(XSetSelectionOwner(display,selection,w,time));
    }
    break;
  case XCConvertSelection: /* Selection */
    {
      
      Atom selection;
      Atom target;
      Atom property;
      Time time = CurrentTime;

      p = evalArgs(EARG1(p),3);
      selection = ATOMOF(getpart(p,0,3));
      target = ATOMOF(getpart(p,1,3));
      property = ATOMOF(getpart(p,2,3));
      if (debug) fprintf(stderr, "XConvertSelection(%lx,%lx,%lx,%lx,%lx)\n",(Int)display,selection,target,property,wi);
      WHEN_NOT_REPLAY(XConvertSelection(display,selection,target,property,wi,time));
    }
    break;
  case XCChangeProperty : /*
    Window
    Atom -- property
    Atom -- type
    Int -- format
    PropertyMode
    String */
    {
      Window w;
      Atom property;
      Atom type;
      int format;
      int mode;
      char buf[100000];

      p = evalArgs(p,6);
      w = INTOF(getpart(p,0,6));
      property = ATOMOF(getpart(p,1,6));
      type = ATOMOF(getpart(p,2,6));
      format = INTOF(getpart(p,3,6));
      mode = INTOF(getpart(p,4,6));
      evalstring(getpart(p,5,6),buf,sizeof buf);
      if(debug) fprintf(stderr,"XChangeProperty(%lx,%lx,%lx,%lx,%d,%d)\n",Idisplay,w,property,type,format,mode);
      if(debug>1) fprintf(stderr,"XChangeProperty: \"%s\"\n",buf);
      WHEN_NOT_REPLAY(XChangeProperty(display,w,property,type,format,mode,buf,strlen(buf)));
    }
    break;
  case XCFreeColors : /* ColormapId [Pixel] Pixel */
    {
      Colormap cmap;
      unsigned long planes;
      unsigned long pixels[10000];
      PTR list;
      int npixels;

      p = evalArgs(p,3);
      cmap = INTOF(getpart(p,0,3));
    WHEN_NOT_REPLAY_START
      if(!cmap) {
	cmap=DefaultColormap(display,DefaultScreen(display));
      }
    WHEN_NOT_REPLAY_END
      list = getpart(p,1,3);
      planes = INTOF(getpart(p,2,3));
      for(npixels=0;!ISNIL(list);npixels++,list=ETLOF(list))
	SAVE(list,pixels[npixels]=INTOF(EHDOF(list)));
      if(debug) fprintf(stderr,"XFreeColors(colormap=%lx,npixels=%d,planes=%lx)\n",cmap,npixels,planes);
      WHEN_NOT_REPLAY(XFreeColors(display,cmap,pixels,npixels,planes));
    }
    break;
  case XCReparentWindow : /* Window */
    {
      Window newparent,child;
      int x,y;
      p = evalArgs(p,1);
      newparent = INTOF(getpart(p,0,1));
    WHEN_NOT_REPLAY_START
      if(!newparent) {
	newparent = DefaultRootWindow(display);
      }
      if(!XTranslateCoordinates(display,wi,newparent,0,0,&x,&y,&child)) {
	fprintf(stderr,"XTranslateCoordinates: windows on different screens!\n");
	exit(1);
      }
      if(debug) fprintf(stderr,
			"XReparentWindow(disp=%lx,window=%lx,parent=%lx,x=%d,y=%d)\n",
			Idisplay,wi,newparent,x,y);
      XReparentWindow(display,wi,newparent,x,y);
    WHEN_NOT_REPLAY_END
    }
    break;
  case XCWarpPointer:	/* tPoint */
    {
      int x, y;
      
      getpoint(EARG1(p), &x, &y);
      if (debug) fprintf(stderr, "XWarpPointer(display=%lx,window=%lx,x=%d,y=%d)\n",Idisplay,wi,x,y);
      WHEN_NOT_REPLAY(XWarpPointer(display,None,wi,0,0,0,0,x,y));
    }
    break;
  case XCSetRegion: /* GCId Rect */
    { GC gc;
      Region r;
      XRectangle b;
      int x,y,w,h;
     
      p = evalArgs(p,2);
      SAVE(p,getrect(ARG2(p,2),&x,&y,&w,&h));
      b.x=x,b.y=y,b.width=w,b.height=h; /* types...*/
      getgc(ARG2(p,1),&gc);
    WHEN_NOT_REPLAY_START
      r=XCreateRegion();
      XUnionRectWithRegion(&b,r,r);
      if (debug) fprintf(stderr, "XSetRegin(%lx,%lx,r), r=rect{%d,%d,%d,%d}\n", (Int)Idisplay, (Int)gc,b.x,b.y,b.width,b.height);
      XSetRegion(display,gc,r);
      XDestroyRegion(r);
    WHEN_NOT_REPLAY_END
    }
    break;
  case XCAddToSaveSet: /*  */
    { 
      if (debug) fprintf(stderr, "XAddToSaveSet(%lx,%lx)\n", Idisplay, wi);
      WHEN_NOT_REPLAY(XAddToSaveSet(display,wi));
    }
    break;
  case XCRemoveFromSaveSet: /*  */
    { 
      if (debug) fprintf(stderr, "XRemoveFromSaveSet(%lx,%lx)\n", Idisplay, wi);
      WHEN_NOT_REPLAY(XRemoveFromSaveSet(display,wi));
    }
    break;
  case XCBell: /* Int */
    {
      int percent = INTOF(EARG1(p));
      if(debug) fprintf(stderr,"XBell(%lx,%d)\n",Idisplay,percent);
      WHEN_NOT_REPLAY(XBell(display,percent));
    }
    break;
  case XCSetGCWarningHack: /* { gcon,gcoff::PixmapId } */
    p=evalArgs(p,2);
#ifdef PROFILE
    if(!replay)
#endif
      {
	Pixmap gcon,gcoff;
#ifdef HBC
	if(!gcwarning.display) add_gc_hook(gcwarninghook);
#endif
	gcwarning.display=display;
	gcwarning.window=wi;
	gcwarning.pixmap[0]=gcon=INTOF(getpart(p,0,2));
	gcwarning.pixmap[1]=gcoff=INTOF(getpart(p,1,2));
	if(debug) fprintf(stderr,"XCSetGCWarningHack %lx %lx\n",gcon,gcoff);
      }
    break;
  default:
    /*if (debug)*/ fprintf(stderr, "Unimplemented X Command %d\n", getcno(p));
  }
}

#define MkXResp(cno,arg) mkconstr1(cno,arg)
#define MkIntXResp(cno,arg) MkXResp(cno,mkint(arg))
#define MkPtrXResp(cno,arg) MkIntXResp(cno,(Int)arg)

static PTR
MkChildList(windows,n)
     Window *windows;
     unsigned int n;
{
  PTR p;
  int i;

  p=mknil();
  for(i=n-1;i>=0;i--) {
    Window w;
    GCSAVE(p);
    REPLAY_ONE_AND_SKIP(w)
      w=windows[i];
    RECORD_ONE(w);
    p=mkcons(mkint(w),p);
  }
  return p;
}

static PTR
MkBitmapRead(w,h,x,y,p)
     int w,h,x,y;
     Pixmap p;
{
  if (debug)
    fprintf(stderr, "BitmapReturn %d %d %d %d %lx\n", w, h, x, y, (long)p);
  return MkXResp(XRBitmapRead,
		 mkconstr(1, 3,
			  mkPoint(w,h),
			  x==-1 ? mkNone()
			        : mkSome(mkPoint(x,y)),
			  mkint(p)));
}

static PTR
mkVisual(v)
     Visual *v;
{
  return mkconstr(0,7,
		  mkint(v->visualid),
		  mkconstr0(v->class),
		  mkint(v->red_mask),
		  mkint(v->green_mask),
		  mkint(v->blue_mask),
		  mkint(v->bits_per_rgb),
		  mkint(v->map_entries));
}

#define OUTOFORDER \
    fprintf(stderr,"Request %d temporarily out of order\n",getcno(p)); \
    exit(1);

PTR
xdorequest(display,wi,p)
     Display *display;
     Window wi;
     PTR p;
{
  /* Variables used in Create{Root,Simple}Window: */
  PTR rp;
  Window parent;
  char resname[1000];

  if (debug > 1) fprintf(stderr, "xdorequest %d\n", getcno(p));
  switch(getcno(p)) {
  case XRqOpenDisplay: /* DisplayName */
    {
      char displayname[1000];
      Display *display;
      
      inittimeq(); /* another side effect... */
      evalstring(EARG1(p), displayname, sizeof displayname);
      if(debug) fprintf(stderr,"XOpenDisplay(\"%s\")\n",displayname);
      REPLAY_ONE_AND_SKIP(display)
	display=XOpenDisplay(displayname[0] ? NULL : displayname);
      RECORD_ONE(display);
      if(debug) fprintf(stderr,"display: %lx\n",(Int)display);
      return MkPtrXResp(XRDisplayOpened,display);
    }
    break;
  case XRqCreateRootWindow:	/* Rect String */
    REPLAY_ONE_AND_SKIP(parent)
      parent = DefaultRootWindow(display);
    RECORD_ONE(parent);
    p=evalArgs(p,2);
    rp = ARG2(p,1);
    evalstring(ARG2(p,2),resname,sizeof resname);
    goto cre;
    /* fall into */
  case XRqCreateSimpleWindow:	/* *w tRect */
    {
      Window this;
      /* XSetWindowAttributes attr;*/
      unsigned int x,y,width,height,border_width;
      unsigned long int border, background;
      int screen;
      
      parent = wi;
      rp = evaluate(ARG2(p, 2));
    cre:
      getrect(rp, &x,&y,&width,&height);
      REPLAY_ONE_AND_SKIP(this) {
	screen=DefaultScreen(display); /* which screen?! */
	border_width=0; /* Should agree with border_width in WindowF.hs */
	border=BlackPixel(display,screen);
	background=WhitePixel(display,screen);
	if (debug) fprintf(stderr, "CreateSimpleWindow(%lx,%lx,%d,%d,%d,%d,%d,%ld,%ld)\n", Idisplay,parent,x,y,width,height,
			   border_width,border,background);
	this=XCreateSimpleWindow(display,parent,x,y,width,height,border_width,border,background);
	if(getcno(p)==XRqCreateRootWindow) {
	  XClassHint ch = { resname,"Fudgets" };
	  XStoreName(display, this, "Fudgets");
	  XSetClassHint(display, this, &ch);
	  /* Program should be allowed to do this too. */
	}
      }
      RECORD_ONE(this);
      return MkIntXResp(XRWindowCreated,this);
    }
    break;
  case XRqCreateGC:		/* Drawable tGCId tGCAttributeList */
    {
      GC gc, oldgc;
      Drawable d;
      XGCValues values;
      unsigned long valuemask;
      
      p = evalArgs(p,3);

      REPLAY_ONE_AND_SKIP(wi)
	if(wi==-1) {   /* To be on the safe side */
	  wi=DefaultRootWindow(display);
	}
      RECORD_ONE(wi);
      if (debug) fprintf(stderr, "CreateGC p = %lx\n", (Int)p);
      SAVE(p,d=getdrawable(wi,getpart(p,0,3)));
      if (debug) fprintf(stderr, "CreateGC p' = %lx\n", (Int)p);
      getgc(getpart(p,1,3), &oldgc);
      getGCvalues(getpart(p,2,3),&values,&valuemask);
      
      if (debug) fprintf(stderr, "CreateGC %lx %lx\n", d, (long)oldgc);

      REPLAY_ONE_AND_SKIP(gc) {
	if(oldgc==MyDefaultGC) 
	  gc=defaultGC(display,d); /* no evaluatin */
	else {
	  gc=XCreateGC(display,d,0,NULL);
	  XCopyGC(display,oldgc,GCAllComponents,gc);
	}
	XChangeGC(display,gc,valuemask,&values);
      }
      RECORD_ONE(gc);
      return MkPtrXResp(XRGCCreated, gc);
    }
    break;
  case XRqLoadFont:		/* String */
    {
      Font fid;
      char fname[1000];
      
      evalstring(EARG1(p), fname, sizeof fname);
      
      if (debug) fprintf(stderr, "XLoadFont(%lx,\"%s\")\n", Idisplay, fname);
      REPLAY_ONE_AND_SKIP(fid)
	fid = XLoadFont(display, fname);
      RECORD_ONE(fid);
      return MkIntXResp(XRFontLoaded, fid);
    }
    break;
  case XRqCreateFontCursor:	/* Int */
    {
      int shape;
      Cursor cursor;
      
      shape = INTOF(EARG1(p));
      if (debug) fprintf(stderr, "XCreateFontCursor(%lx,%d)\n", Idisplay,shape);
      REPLAY_ONE_AND_SKIP(cursor)
	cursor=XCreateFontCursor(display,shape);
      RECORD_ONE(cursor);
      return MkIntXResp(XRCursorCreated, cursor);
    }
    break;
  case XRqGrabPointer:		/* Bool [EventMask] */
    {
      int result;
      Window confine_to=None;
      Bool owner_events;
      unsigned int event_mask=DEFAULT_GRAB_MASK;
      int pointer_mode=GrabModeAsync,keyboard_mode=GrabModeAsync;
      Cursor cursor=None;
      Time time=CurrentTime;
      
      p = evalArgs(p,2);
      owner_events = GET_BOOL_VAL(ARG2(p,1));
      event_mask = geteventmask(ARG2(p,2));
      if (debug) fprintf(stderr, "XGrabPointer(%lx,%lx,%d,%x,%d,%d,%lx,%lx,%lx)\n",Idisplay,wi,owner_events,event_mask,pointer_mode,
			 keyboard_mode,confine_to,cursor,time);
      REPLAY_ONE_AND_SKIP(result)
	result=XGrabPointer(display,wi,owner_events,event_mask,pointer_mode,
			    keyboard_mode,confine_to,cursor,time);
      RECORD_ONE(result);
      return MkIntXResp(XRPointerGrabbed, result);
    }
    break;
  case XRqQueryFont:		/* FontId */
#if 1
    OUTOFORDER
#else
    /* Not used at the moment. */
    /* Not prepared for two-pass profiling ! */
    {
      Font fid;
      XFontStruct *fs;
      
      fid = INTOF(EARG1(p));
      if (debug) fprintf(stderr, "XQueryFont(%lx,%lx)\n", Idisplay, fid);
      fs=XQueryFont(display, fid);
      return MkXResp(XRFontQueried, mkFontStruct(fs));
    }
    break;
#endif
  case XRqQueryFontList:		/* FontId */
    {
      Font fid;
      XFontStruct *fs;
      PTR ptrfs;

      fid = INTOF(EARG1(p));
      if (debug) fprintf(stderr, "XQueryFont(%lx,%lx)\n", Idisplay, fid);
      REPLAY_ONE_AND_SKIP(fs)
	fs=XQueryFont(display, fid);
      RECORD_ONE(fs);
      ptrfs = mkFontStructListMaybe(fs);  /* Knows about two-pass profiling */
      WHEN_NOT_REPLAY(if(fs) XFreeFontInfo(NULL, fs, 1));
      return MkXResp(XRFontQueriedList, ptrfs);
    }
    break;
  case XRqLoadQueryFontList: /* String */
    {
      char fname[1000];
      XFontStruct *fs;
      PTR ptrfs;

      evalstring(EARG1(p), fname, sizeof fname);
      
      if (debug) fprintf(stderr, "XLoadQueryFont(%lx,\"%s\")\n", Idisplay, fname);
      REPLAY_ONE_AND_SKIP(fs)
	fs = XLoadQueryFont(display, fname);
      RECORD_ONE(fs);
      ptrfs = mkFontStructListMaybe(fs);  /* Knows about two-pass profiling */
      if(debug) fprintf(stderr, "%lx = mkFontStructListMaybe(%lx)\n", (Int)ptrfs,(Int)fs);
      WHEN_NOT_REPLAY(if(fs) XFreeFontInfo(NULL, fs, 1));
      return MkXResp(XRFontQueriedList, ptrfs);
    }
    break;
  case XRqAllocNamedColor:	/* ColormapId ColorName */
    {
      Colormap colormap;
      char colorname[1000];
      XColor ecolor,color;
      int screen;
      int bool;
      WHEN_NOT_REPLAY(screen = DefaultScreen(display)); /* which screen?! */
      
      p = evalArgs(p,2);
      colormap = INTOF(ARG2(p,1));
      WHEN_NOT_REPLAY(if(!colormap) colormap=DefaultColormap(display,screen));
      evalstring(ARG2(p,2), colorname, sizeof colorname);
      if (debug) fprintf(stderr, "XAllocNamedColor(%lx,%s)\n", colormap,colorname);
#ifdef PROFILE
      if(replay) {
	REPLAY_BOOL(bool);
	if(bool) REPLAY(color);
      } else
#endif
	bool = XAllocNamedColor(display,colormap,colorname,&color,&ecolor);
#ifdef PROFILE
      if(record) {
	RECORD_BOOL(bool);
	if(bool)
	  RECORD(color);
      }
#endif
      if(bool)
	return MkXResp(XRColorAllocated, mkSome(mkColor(color)));
      else
	return MkXResp(XRColorAllocated, mkNone());
    }
    break;
  case XRqAllocColor:	/* ColormapId RGB */
    {
      Colormap colormap;
      XColor color;
      int bool,screen;

      WHEN_NOT_REPLAY(screen=DefaultScreen(display)); /* which screen?! */
      
      p = evalArgs(p,2);
      colormap = INTOF(ARG2(p,1));
      if(!colormap) colormap=DefaultColormap(display,screen);
      getRGB(ARG2(p,2),&color.red,&color.green,&color.blue);
      if (debug) fprintf(stderr, "XAllocColor(cmap=%lx,r=%d,g=%d,b=%d)\n", 
			 colormap,color.red,color.green,color.blue);
#ifdef PROFILE
      if(replay) {
	REPLAY_BOOL(bool);
	if(bool) REPLAY(color);
      } else
#endif
	bool = XAllocColor(display,colormap,&color);
#ifdef PROFILE
      if(record) {
	RECORD_BOOL(bool);
	if(bool)
	  RECORD(color);
      }
#endif
      if(bool)
	return MkXResp(XRColorAllocated, mkSome(mkColor(color)));
      else
	return MkXResp(XRColorAllocated, mkNone());
    }
    break;
  case XRqCreatePixmap:	/* Size Depth */
    {
      Pixmap pixmap;
      unsigned int width,height,depth;
      int screen;
      WHEN_NOT_REPLAY(screen=DefaultScreen(display)); /* which screen?! */
      
      p = evalArgs(p,2);
      depth=INTOF(ARG2(p,2));
      getpoint(ARG2(p,1),&width,&height);
      REPLAY_ONE_AND_SKIP(pixmap) {
	if(depth==CopyFromParent) depth=DefaultDepth(display,screen);
	if (debug) fprintf(stderr, "XCreatePixmap(%lx,%lx,%d,%d,%d)\n", Idisplay,wi,width,height,depth);
	if(wi==-1) wi=DefaultRootWindow(display);
	pixmap=XCreatePixmap(display,wi,width,height,depth);
	RECORD_ONE(pixmap);
      }
      return MkIntXResp(XRPixmapCreated, pixmap);
    }
    break;
  case XRqReadBitmapFile:   /* name */
    {
      char filename[2000];
      unsigned int width, height;
      Pixmap bitmap;
      Window w;
      int x_hot, y_hot;
      int r;
      WHEN_NOT_REPLAY(w = DefaultRootWindow(display));
      evalstring(EARG1(p), filename, sizeof filename);
      if (debug) fprintf(stderr, "XReadBitmapFile(%lx,%lx,%s)\n", Idisplay,w,filename);
#ifdef PROFILE
      if(replay) {
	REPLAY(r);
	REPLAY(width);
	REPLAY(height);
	REPLAY(bitmap);
	REPLAY(x_hot);
	REPLAY(y_hot);
      } else 
#endif
	r = XReadBitmapFile(display, w, filename,
			    &width, &height, &bitmap, &x_hot, &y_hot);
#ifdef PROFILE
      if(record) {
	RECORD(r);
	RECORD(width);
	RECORD(height);
	RECORD(bitmap);
	RECORD(x_hot);
	RECORD(y_hot);
      }
#endif
      if (r == BitmapSuccess) {
	if (debug) fprintf(stderr, "success\n");
	return MkBitmapRead(width,height,x_hot,y_hot,bitmap); /* Knows about two-pass profiling */
      } else {
	if (debug) fprintf(stderr, "fail %d\n", r);
	return MkXResp(XRBitmapRead, mkconstr0(0));
      }
    }
    break;
  case XRqCreateBitmapFromData:     /* size, optional hotspot, data */
#if 1
    OUTOFORDER
#else
    {  /* Not prepared for two-pass profiling */
      char data[10000];
      unsigned int width, height;
      Pixmap r;
      int x_hot, y_hot;
      PTR q = ARG1(p);
      
	getpoint(getpart(q,0,3),&width,&height);
      if (getcno(getpart(q,1,3)) == 0) {
	x_hot = y_hot = -1;
      } else {
	getpoint(getpart(getpart(q,1,3), 1, 2), &x_hot, &y_hot);
      }
      tocstring(getpart(q,2,3), data, sizeof data);
      if (debug) fprintf(stderr, "XCreateBitmapFromData(%lx,%lx,%lx,%d,%d)\n",Idisplay,DefaultRootWindow(display),(long)data,width,height);
      r = XCreateBitmapFromData(display, wi, data, width, height);
      addxidresult(XRBitmapReturn, wi, r);
      addnumresult(-1, width, height);
      addnumresult(-1, x_hot, y_hot);
    }
#endif
    break;
  case XRqRmGetStringDatabase: /* String */
    {
      char buf[100000]; /* Urrk! */
      evalstring(EARG1(p),buf, sizeof buf); /* Knows about two-pass profiling */
      if (debug) fprintf(stderr, "XrmGetStringDatabase(\"%s\")\n",buf);
      return MkPtrXResp(XRRmDatabaseCreated, XrmGetStringDatabase(buf));
      /* num?! */
    }
    break;
  case XRqRmGetResource: /* RmDatabase String String */
    OUTOFORDER
#if 0
    {
      XrmDatabase db;
      char name[100],class[100],*type;
      XrmValue value;
      Bool gotit;
      db = (XrmDatabase)INTOF(getpart(p,0,3));
      tocstring(getpart(p,1,3), name, sizeof name);
      tocstring(getpart(p,2,3), class, sizeof class);
      if (debug) fprintf(stderr, "XrmGetResource(%lx,\"%s\",\"%s\")\n",(long)db,name,class);
      gotit = XrmGetResource(db, name, class, &type, &value);
      addnumresult(XRGotResource, wi, gotit);
      if(gotit) {
	addptrresult((long)type, value.size, value.addr);
      }
    }
#endif
    break;
  case XRqTranslateCoordinates:
    {
      int dx,dy;
      Window child;
      if(debug) fprintf(stderr,"XTranslateCoordinates(%lx,%lx,%lx,...)\n",Idisplay,wi,DefaultRootWindow(display));
#ifdef PROFILE
      if(replay) {
	REPLAY(dx);
	REPLAY(dy);
      }
#endif
      (void)XTranslateCoordinates(display,wi,DefaultRootWindow(display),0,0,&dx,&dy,&child);
#ifdef PROFILE
      if(record) {
	RECORD(dx);
	RECORD(dy);
      }
#endif
      return MkXResp(XRCoordinatesTranslated,mkPoint(dx,dy));
    }
    break;
  case XRqInternAtom: /* String Bool */
    {
      char buf[1000];
      int if_exists;
      Atom atom;
      
      p = evalArgs(p,2);
      if_exists = GET_BOOL_VAL(ARG2(p,2));
      evalstring(ARG2(p,1), buf, sizeof buf);
      if (debug) fprintf(stderr, "XInternAtom(%lx,\"%s\",%d)\n",Idisplay, buf, if_exists);
      REPLAY_ONE_AND_SKIP(atom)
	atom=XInternAtom(display, buf, if_exists);
      RECORD_ONE(atom);
      return MkXResp(XRGotAtom,mkAtom(atom));
    }
  case XRqGetWindowProperty: /*
    Int -- offset
    Atom -- property
    Bool -- delete
    Atom -- req_type */
    {
#define PROPBUF 4000
      long offset;
      long length = PROPBUF/4;
      Atom property;
      Bool delete;
      Atom req_type;
      PTR str;

      Atom actual_type;
      int actual_format;
      unsigned long nitems,bytes_after;
      unsigned char* prop;

      p = evalArgs(p,4);
      offset = INTOF(getpart(p,0,4));
      property = ATOMOF(getpart(p,1,4));
      delete = GET_BOOL_VAL(getpart(p,2,4));
      req_type = ATOMOF(getpart(p,3,4));
      if(debug) 
	fprintf(stderr,"XGetWindowProperty(%lx,%lx,%lx,%ld,%ld,%d,%lx)\n",
		Idisplay,wi,property,offset,length,delete,req_type);

#ifdef PROFILE
      if(replay) {
	REPLAY(actual_type);
	REPLAY(actual_format);
	REPLAY(nitems);
      } else
#endif
	{
	  int s = XGetWindowProperty(display,wi,property,offset,length,
				     delete,req_type,&actual_type,&actual_format,
				     &nitems,&bytes_after,&prop);
	  if(s != Success) {
	    fprintf(stderr,"XGetWindowProperty returned %d\n",s);
	  }
	}
#ifdef PROFILE
      if(record) {
	RECORD(actual_type);
	RECORD(actual_format);
	RECORD(nitems);
      }
#endif
      if(actual_format != None) {
	str = mkestring(prop);  /* Knows about two-pass profiling */
	WHEN_NOT_REPLAY(XFree(prop));
      } else
	str = mknil();

      return mkconstr(XRGotWindowProperty,5,
		      mkAtom(actual_type),
		      mkint(actual_format),
		      mkint(nitems),mkint(bytes_after),str);
    }
  case XRqQueryPointer:
    { Window root,child;
      int root_x,root_y,win_x,win_y;
      unsigned int mask;
      Bool same_screen;

#ifdef PROFILE
      if(replay) {
	REPLAY(same_screen);
	REPLAY(root_x);
	REPLAY(root_y);
	REPLAY(win_x);
	REPLAY(win_y);
	REPLAY(mask);
      } else
#endif
	same_screen = XQueryPointer(display,wi,&root,&child,
				    &root_x,&root_y,&win_x,&win_y,
				    &mask);
#ifdef PROFILE
      if(record) {
	RECORD(same_screen);
	RECORD(root_x);
	RECORD(root_y);
	RECORD(win_x);
	RECORD(win_y);
	RECORD(mask);
      }
#endif	
      return mkconstr(XRPointerQueried,4,
		      mkBool(same_screen),
		      mkPoint(root_x,root_y),mkPoint(win_x,win_y),
		      mkModState(mask));
    }
  case XRqQueryColor: /* ColormapId Pixel */
    { XColor color;
      Colormap cmap;

      p = evalArgs(p,2);
      cmap = INTOF(getpart(p,0,2));
      REPLAY_ONE_AND_SKIP(color) {
	if(!cmap) cmap=DefaultColormap(display,DefaultScreen(display));
	color.pixel = INTOF(getpart(p,1,2));
	if(debug) 
	  fprintf(stderr,"XQueryColor(%lx,%lx,{%d})\n",(Int)Idisplay,(Int)cmap,(int)color.pixel);
	XQueryColor(display,cmap,&color);
      } 
      RECORD_ONE(color);
      return mkconstr1(XRColorQueried,mkColor(color));
    }
  case XRqQueryTree: /* void */
    { Status s;
      Window root,parent,*children;
      unsigned int n_children;
      PTR child_list;

      if(debug) fprintf(stderr,"XQueryTree(%lx,%lx,...)\n",Idisplay,wi);
#ifdef PROFILE
      if(replay) {
	REPLAY(s);
	REPLAY(n_children);
      } else 
#endif
	s=XQueryTree(display,wi,&root,&parent,&children,&n_children);
      if(!s) {
	fprintf(stderr,"XQueryTree failed!\n");
	exit(1);
	/* !! return error response instead! */
      }
#ifdef PROFILE
      if(record) {
	RECORD(s);
	RECORD(n_children);
      }
#endif
      child_list=MkChildList(children,n_children); /* Knows about two pass profiling */
      WHEN_NOT_REPLAY(XFree(children));
      return mkconstr(XRTreeQueried,3,mkint(root),mkint(parent),child_list);
    }
  case XRqDefaultRootWindow: /* void */
    { Window root;

      if(debug) fprintf(stderr,"DefaultRootWindow(%lx)\n",Idisplay);
      REPLAY_ONE_AND_SKIP(root)
	root=DefaultRootWindow(display);
      RECORD_ONE(root);
      return mkconstr1(XRGotDefaultRootWindow,mkint(root));
    }
  case XRqGetGeometry: /* void */
    { Window root;
      int x,y;
      unsigned int w,h,bw,depth;

      if(debug) fprintf(stderr,"GetGeometry(%lx)\n",Idisplay);
#ifdef PROFILE
      if(replay) {
	REPLAY(x);
	REPLAY(y);
	REPLAY(w);
	REPLAY(h);
	REPLAY(bw);
	REPLAY(depth);
      } else
#endif
      XGetGeometry(display,wi,&root,&x,&y,&w,&h,&bw,&depth);
#ifdef PROFILE
      if(record) {
	RECORD(x);
	RECORD(y);
	RECORD(w);
	RECORD(h);
	RECORD(bw);
	RECORD(depth);
      }
#endif
      return mkconstr(XRGotGeometry,3,mkRect(x,y,w,h),mkint(bw),mkint(depth));
    }
  case XRqDefaultVisual: /* void */
    { int screen=DefaultScreen(display);
      Visual *v;
      if(debug) fprintf(stderr,"DefaultVisual(%lx,%d)\n",Idisplay,screen);
#ifdef PROFILE
      if(replay) {
	Visual rv;
	REPLAY(rv);
	v=&rv;
      }
      else
#endif
	v=DefaultVisual(display,screen);
      RECORD_ONE(*v);
      return MkXResp(XRGotVisual,mkVisual(v));
    }
  case XRqSync: /* Bool */
    {
      Bool discard=GET_BOOL_VAL(EARG1(p));
      if(debug) fprintf(stderr,"XSync(%lx,%d)\n",Idisplay,discard);
      WHEN_NOT_REPLAY(XSync(display,discard));
      return mkconstr0(XRSynced);
    }
  case XRqQueryTextExtents16: /* FontId String */
    {
      Font fid;
      XChar2b buf[5000];
      int length;
      int dir,ascent,descent;
      XCharStruct overall;
      
      p = evalArgs(p,2);
      fid = INTOF(ARG2(p,1));
      length=evalstring16(ARG2(p,2), buf, sizeof buf/sizeof(XChar2b));
      if(debug) fprintf(stderr,"XQueryTextExtents16(%lx,%lx,\"...\",...)\n",
			Idisplay,fid/*,buf*/);
#ifdef PROFILE
      if(replay) {
	REPLAY(dir);
	REPLAY(ascent);
	REPLAY(descent);
	REPLAY(overall);
      } else
#endif
      XQueryTextExtents16(display,fid,buf,length,&dir,&ascent,&descent,&overall);
#ifdef PROFILE
      if(record) {
	RECORD(dir);
	RECORD(ascent);
	RECORD(descent);
	RECORD(overall);
      }
#endif
      return mkconstr(XRTextExtents16Queried,3,
		      mkint(ascent),mkint(descent),
		      mkCharStruct(&overall));
      
    }
  case XRqListFonts: /* String Int -- pattern, maxnames */
    {
      char pattern[256];
      int maxnames,count;
      char **list;
      PTR strlist;

#ifdef PROFILE
      if(record||replay) { OUTOFORDER; }
#endif
      p=evalArgs(p,2);
      maxnames = INTOF(ARG2(p,2));
      (void)evalstring(ARG2(p,1), pattern, sizeof pattern);
      if(debug) fprintf(stderr,"XListFonts(%lx,\"%s\",%d,...)\n",
			Idisplay,pattern,maxnames);
      list=XListFonts(display,pattern,maxnames,&count);
      strlist=mkStringList(list,count);
      XFreeFontNames(list);
      return mkconstr1(XRGotFontList,strlist);
    }
    break;
  case XRqListFontsWithInfo: /* String Int -- pattern, maxnames */
    {
      char pattern[256];
      int maxnames,count;
      char **list;
      XFontStruct *info;
      PTR fontinfolist;

#ifdef PROFILE
      if(record||replay) { OUTOFORDER; }
#endif
      p=evalArgs(p,2);
      maxnames = INTOF(ARG2(p,2));
      (void)evalstring(ARG2(p,1), pattern, sizeof pattern);
      if(debug) fprintf(stderr,"XListFontsWithInfo(%lx,\"%s\",%d,...)\n",
			Idisplay,pattern,maxnames);
      list=XListFontsWithInfo(display,pattern,maxnames,&count,&info);
      fontinfolist=mkFontInfoList(list,info,count);
      XFreeFontInfo(list,info,count);
      if(debug) fprintf(stderr,"XListFontsWithInfo done\n");
      return mkconstr1(XRGotFontListWithInfo,fontinfolist);
    }
    break;
  default:
    /*if (debug)*/ fprintf(stderr, "Unimplemented X Request %d\n", getcno(p));
    exit(1);
  }
}


PTR doxcall(t, p)
     int t; /* tag of the Request */
     PTR p; /* pointer the argument of the Request */
{
    /*int GCs_before;*/
  PTR response;

  if(debug>2) {
    fprintf(stderr,"xhandle before gc\n");
    GCSAVE(p);
  }
  p = evaluate(p);
  switch(t) {
    /* X command and request handling */
  case XCommand: /* (Display, Window, XCommand) */
    p = evalArgs(p,3);
    xdocommand((Display *)INTOF(GET1OF3(p)), INTOF(GET2OF3(p)), GET3OF3(p));
    response=mkconstr0(RSuccess);
    break;
  case XRequest: /* (Display, Window, XRequest) */
    {
      PTR xresp;
      p = evalArgs(p,3);
      xresp=xdorequest((Display *)INTOF(GET1OF3(p)), INTOF(GET2OF3(p)), GET3OF3(p));
      response=mkconstr1(XResponse,xresp);
    }
    break;
    
  default:
    fprintf(stderr, "Unknown X I/O request %d (not %d or %d)\n", t,XCommand,XRequest);
    exit(1);
    break;
  }
  {
#if 0
    if(hp>ehp) {
      fprintf(stderr,"Heap overrun by %d words in xhandle\n",hp-ehp);
    }
#endif
  }
  if(debug>2) {
    fprintf(stderr,"xhandle after gc\n");
    GCSAVE(response);
  }
  return response;
}

extern void mkresponse();

static void
xhandle(t, p)
int t;
PTR p;
{
  mkresponse(doxcall(t,p));
}

extern void shandle PROTO((int,PTR));

void (*xhandler) PROTO((int, PTR)) = xhandle;
void (*shandler) PROTO((int, PTR)) = shandle;
