#include <stdio.h>
#include <errno.h>

#define XLIB_ILLEGAL_ACCESS	/* to access Display struct in Ultrix */
#include <X11/Xlib.h>
#include <X11/Xresource.h>
#include <X11/Xutil.h>
#include <sys/types.h>
#include <sys/time.h>
#include "dialog.h"
#include "heap.h"
#include "xlib.h"
#include "timeq.h"
#include "xencode.h"

static Int rescno;

PTR
mkModState(n)
unsigned int n;
{
    PTR res;
    int i;

    res = mknil();
    for(i=12; i>=0; i--) {
	if (n & (1<<i))
	    res = mkcons(mkenum(i),res);
    }
    return res;
}

PTR
mknstring(s,l)
unsigned char *s;
unsigned int l;
{
  PTR p;
  
  for(s += l, p = mknil(); l > 0; l--) {
    unsigned char c;
    GCSAVE(p);
    REPLAY_ONE_AND_SKIP(c)
      c = *--s;
    RECORD_ONE(c);
    p = mkcons(mkchar(c), p);
  }
  return p;
}

PTR
mkStringList(list,l)
char **list;
unsigned int l;
{
  PTR p,sp;
  
  for(list += l, p = mknil(); l > 0; l--) {
    char *s;
    s = *--list;
    SAVE(p,sp=mkestring(s));
    p = mkcons(sp, p);
  }
  return p;
}

PTR
mkFontInfoList(list,info,l)
char **list;
XFontStruct *info;
unsigned int l;
{
  PTR p,sp,ip;
  
  for(list += l, info += l, p = mknil(); l > 0; l--) {
    PUSHPTR(p);
    sp=mkestring(*--list);
    PUSHPTR(sp);
    ip=mkFontStructList(--info);
    POPPTR(sp);
    POPPTR(p);
#if 1
    p = mkcons(mkpair(sp, ip), p);
#else
    p = mkcons(ip, p);
#endif
  }
  return p;
}

/* CharStruct 5*Int */
PTR
mkCharStruct(cs)
XCharStruct *cs;
{
    if(debug > 1) fprintf(stderr, "mkCharStruct lb %d rb %d width %d ascent %d descent %d\n", cs->lbearing,cs->rbearing,cs->width,cs->ascent,cs->descent);
    return mkconstr(0, 5, mkint(cs->lbearing), mkint(cs->rbearing), mkint(cs->width), mkint(cs->ascent), mkint(cs->descent));
}

/* MAY GARBAGE COLLECT! */
/* FontQueried FontStruct */
/* FontStruct FontId FontDirection Char Char Bool Char (List (String # Int)) CharStruct CharStruct (Option (Array CharStruct)) Int Int */
/*            0      1             2    3    4    5    6                     7          8                  9                   10  xx */
#if 0
/* Not prepared for tw-pass profiling */
PTR
mkFontStruct(fs)
XFontStruct *fs;
{
    PTR res, elem9;
    int arrsize;

    if(!fs) return mkNone();
    arrsize = fs->max_char_or_byte2 - fs->min_char_or_byte2 + 1;
    if(debug) fprintf(stderr, "mkFontStruct %lx %x %x ascent %d descent %d\n",fs->fid,fs->min_char_or_byte2,fs->max_char_or_byte2,
		      fs->ascent,fs->descent);
    if(fs->per_char) {
	int i;
	PTR arr;
	/* !!! Need a GC check here !!! */
	arr = mkarray(fs->min_char_or_byte2, arrsize);
	elem9 = mkSome(arr); /* Some arr */
	for(i = 0; i<arrsize; i++) {
	    if(debug > 1) fprintf(stderr, "char %x ",i + fs->min_char_or_byte2);
	    *arrayref(arr, i) = mkCharStruct(fs->per_char + i);
	}
    } else 
	elem9 = mkNone(); /* None */
    res = mkconstr(0, 12,
		   mkint(fs->fid), /* 0 */
		   mkint(fs->direction), /* 1 */
		   mkint(fs->min_char_or_byte2), /* 2 */
		   mkint(fs->max_char_or_byte2), /* 3 */
		   mkenum(fs->all_chars_exist), /* 4 */
		   mkint(fs->default_char), /* 5 */
		   mknil(),	/* 6 */
		   mkCharStruct(&fs->min_bounds), /* 7 */
		   mkCharStruct(&fs->max_bounds), /* 8 */
		   elem9,	/* 9 */
		   mkint(fs->ascent), /* 10 */
		   mkint(fs->descent));	/* 11 */
    return mkSome(res);
}
#endif

PTR
mkFontStructListMaybe(fs)
XFontStruct *fs;
{
  return fs ? mkSome(mkFontStructList(fs)) : mkNone();
}

PTR
mkFontStructList(fs)
XFontStruct *fs;
{
  PTR res, elem9;
  int arrsize,minchar,maxchar;

#if 1
  /* fs is handled by the caller, so this should not be needed ??! */
#ifdef PROFILE
  {
    XFontStruct fsbuf;
    if(replay) {
      REPLAY(fsbuf);
      fs = &fsbuf;
    }
    if(record) {
      RECORD(*fs);
    }
  }
#endif
#endif

  /* !!! Passing only minchar and maxchar to the Haskell code is not enough,
     since fs->{min,max}_char_or_byte2 can be >=256 if fs->{min,max}_byte1
     are both 0. */
  minchar = fs->min_char_or_byte2 + 256*fs->min_byte1;
  maxchar = fs->max_char_or_byte2 + 256*fs->max_byte1;
  /* Wrong: arrsize = maxchar - minchar + 1; */
  arrsize = (fs->max_char_or_byte2-fs->min_char_or_byte2+1)*
            (fs->max_byte1-fs->min_byte1+1);
  if(debug) fprintf(stderr, "mkFontStructList %lx %x %x ascent %d descent %d\n",
		    fs->fid,minchar,maxchar,fs->ascent,fs->descent);
  if(fs->per_char) {
    int i;
    elem9 = mknil();
    for(i = arrsize-1; i>=0; i--) {
      XCharStruct *cs,csbuf;
      cs = &csbuf;  /* Only needed if PROFILE */
      GCSAVE(elem9);
      REPLAY_ONE_AND_SKIP(csbuf)
	cs = fs->per_char + i;
      RECORD_ONE(*cs);
      if(debug > 1) fprintf(stderr, "char %x ",i + minchar);
      elem9 = mkcons(mkCharStruct(cs),elem9);
    }
    elem9 = mkSome(elem9);
  } else 
    elem9 = mkNone();
  res = mkconstr(0, 12,
		 mkint(fs->fid), /* 0 */
		 mkint(fs->direction), /* 1 */
		 mkint(minchar), /* 2 */
		 mkint(maxchar), /* 3 */
		 mkenum(fs->all_chars_exist), /* 4 */
		 mkint(fs->default_char), /* 5 */
		 mknil(),	/* 6 */
		 mkCharStruct(&fs->min_bounds), /* 7 */
		 mkCharStruct(&fs->max_bounds), /* 8 */
		 elem9,	/* 9 */
		 mkint(fs->ascent), /* 10 */
		 mkint(fs->descent));	/* 11 */
  return res;
}

/* KeyEvent Time Point Point State Pressed KeyCode KeySym KeyLookup */
/*          Int  (,)    (,)  ??    Bool    Int     String String */
static PTR
putKeyEvent(event)
XEvent *event;
{
    KeySym k;
    unsigned char buf[100];
    int n;
    char *key;
    PTR r;

    n=XLookupString(&event->xkey,(char *)buf,sizeof(buf)-1,&k,NULL);
    if (n >= 0)
	buf[n] = 0;
    if(!(key=XKeysymToString(k))) key="(undefined)";
    r = mkconstr(rescno, 8,
		 mkint(event->xkey.time),
		 mkPoint(event->xkey.x,event->xkey.y),
		 mkPoint(event->xkey.x_root,event->xkey.y_root),
		 mkModState(event->xkey.state),
		 mkenum(event->type == KeyRelease),
		 mkint(event->xkey.keycode),
		 mkestring(key),
		 mkestring(buf));
    if (debug) fprintf(stderr, ", state=%x, x=%d, y=%d, Key%s %d '%s' '%s'", event->xkey.state, event->xkey.x,event->xkey.y,
		       event->type == KeyRelease ? "Up" : "Down", event->xkey.keycode, key, buf);
    return r;
}

/* ButtonEvent Time Point Point State Pressed Button */
static PTR
putButtonEvent(event)
XEvent *event;
{
    PTR r;

    r = mkconstr(rescno, 6,
		 mkint(event->xbutton.time),
		 mkPoint(event->xbutton.x,event->xbutton.y),
		 mkPoint(event->xbutton.x_root,event->xbutton.y_root),
		 mkModState(event->xbutton.state),
		 mkenum(event->type == ButtonRelease),
		 mkconstr1(1, mkint(event->xbutton.button)));
    if (debug) fprintf(stderr, ", state=%x, x=%d, y=%d, Button%s", event->xbutton.state, event->xbutton.x,event->xbutton.y,
		       event->type == ButtonRelease ? "Up" : "Down");
    return r;
}

/* MotionNotify tPoint tPoint tState */
static PTR
putMotionEvent(event)
XEvent *event;
{
    PTR r;

    r = mkconstr(rescno, 4, 
		 mkint(event->xmotion.time),
		 mkPoint(event->xmotion.x,event->xmotion.y),
		 mkPoint(event->xmotion.x_root,event->xmotion.y_root),
		 mkModState(event->xmotion.state));
    if (debug) fprintf(stderr, ", state=%x, x=%d, y=%d, Motion ", event->xmotion.state,event->xmotion.x,event->xmotion.y);
    return r;
}

/* ResizeRequest tPoint */
static PTR
putResizeRequestEvent(event)
XEvent *event;
{
    return (mkconstr1(rescno, mkPoint(event->xresizerequest.width,event->xresizerequest.height)));
}

/* Expose Rect Int */
static PTR
putExposeEvent(event)
XEvent *event;
{
#ifdef HBC
/* Old version. This works, although it is wrong. */
    return (mkconstr1(rescno,
		     mkpair(mkRect(event->xexpose.x,
				   event->xexpose.y,
				   event->xexpose.width,
				   event->xexpose.height),
			    mkint(event->xexpose.count))));
#else
/* New version. This doesn't work with HBC. Why not?!? */
    return (mkconstr(rescno, 2,
		     mkRect(event->xexpose.x,
			    event->xexpose.y,
			    event->xexpose.width,
			    event->xexpose.height),
		     mkint(event->xexpose.count)));
#endif
}

/* GraphicsExpose Rect Int Int Int */
static PTR
putGraphicsExposeEvent(event)
XEvent *event;
{
    return (mkconstr(rescno, 4,
		     mkRect(event->xgraphicsexpose.x,
			    event->xgraphicsexpose.y,
			    event->xgraphicsexpose.width,
			    event->xgraphicsexpose.height),
		     mkint(event->xgraphicsexpose.count),
		     mkint(event->xgraphicsexpose.major_code),
		     mkint(event->xgraphicsexpose.minor_code)));
}

/* EnterNotify tPoint tPoint tDetail tMode Bool */
/* LeaveNotify tPoint tPoint tDetail tMode Bool */
static PTR
putCrossingEvent(event)
XEvent *event;
{
    PTR r;

    r = mkconstr(rescno, 6,
		 mkint(event->xcrossing.time),
		 mkPoint(event->xcrossing.x,event->xcrossing.y),
		 mkPoint(event->xcrossing.x_root,event->xcrossing.y_root),
		 mkenum(event->xcrossing.detail),
		 mkenum(event->xcrossing.mode),
                 mkenum(event->xcrossing.focus));
    if(debug) fprintf(stderr, ", detail = %x, mode = %x", event->xcrossing.detail, event->xcrossing.mode);
    return r;
}

/* FocusIn Detail Mode */
/* FocusOut Detail Mode */
static PTR
putFocusChangeEvent(event)
XEvent *event;
{
    PTR r;

    r = mkconstr(rescno, 2,
		 mkenum(event->xfocus.detail),
		 mkenum(event->xfocus.mode));
    if(debug) fprintf(stderr, ", detail = %x, mode = %x", event->xfocus.detail, event->xfocus.mode);
    return r;
}

static PTR
putVisibilityEvent(event)
XEvent *event;
{
/* ??? printf("%s",visibilityState(event->xvisibility.state)); */
/*    fprintf(stderr, "visibility not implemented\n");
    exit(1); */
    return mkconstr1(rescno, mkenum(event->xvisibility.state));
}

/* ConfigureNotify Rect Int */
static PTR
putConfigureEvent(event)
XEvent *event;
{
#ifdef HBC
/* Old version. This works, although it is wrong. */
    return mkconstr1(rescno,
		     mkpair(mkRect(event->xconfigure.x,event->xconfigure.y,event->xconfigure.width,event->xconfigure.height),
				  mkint(event->xconfigure.border_width)));
#else
    return mkconstr(rescno,2,
		    mkRect(event->xconfigure.x,event->xconfigure.y,event->xconfigure.width,event->xconfigure.height),
		    mkint(event->xconfigure.border_width));
#endif
}

static PTR mkLongList PROTO ((long l[],int n));
static PTR mkShortList PROTO ((short s[],int n));
static PTR mkByteList PROTO ((char b[],int n));

static PTR
mkLongList(l,n)
long l[];
int n;
{
  if(n==0) return mknil();
  else return mkcons(mkint(l[0]),mkLongList(l+1,n-1));
}

static PTR
mkShortList(s,n)
short s[];
int n;
{
  if(n==0) return mknil();
  else return mkcons(mkint(s[0]),mkShortList(s+1,n-1));
}

static PTR
mkByteList(b,n)
char b[];
int n;
{
  if(n==0) return mknil();
  else return mkcons(mkchar(b[0]),mkByteList(b+1,n-1));
}

typedef enum { CDByte, CDShort, CDLong } ClientDataCon;

static PTR
mkClientData(ce)
XClientMessageEvent *ce;
{
  if(debug>1) fprintf(stderr," format=%d\n",ce->format);
  switch(ce->format) {
  case 32:
    return mkconstr1(CDLong,mkLongList(ce->data.l,5));
  case 16:
    return mkconstr1(CDShort,mkShortList(ce->data.s,10));
  case  8:
  default:
    return mkconstr1(CDByte,mkByteList(ce->data.b,20));
  }
}

static PTR 
mkSelection(selection,target,property)
Atom selection,target,property;
{
  return mkconstr(0,3,mkAtom(selection),mkAtom(target),mkAtom(property));
}

/* ClientMessage Atom ClientData */
static PTR
putClientMessageEvent(event)
XEvent *event;
{
#ifdef HBC
  return mkconstr1(rescno,
		  mkpair(mkAtom(event->xclient.message_type),
			 mkClientData(&event->xclient)));
#else
  return mkconstr(rescno,2,
		  mkAtom(event->xclient.message_type),
		  mkClientData(&event->xclient));
#endif
}

/* Atom */
static PTR
putSelectionClearEvent(event)
XEvent *event;
{
  return mkconstr1(rescno, mkAtom(event->xselectionclear.selection));
}

/* Window Selection */
static PTR
putSelectionRequestEvent(event)
XEvent *event;
{
#if 0 /* defined(HBC) */
  return mkconstr1(rescno,
		   mkpair(mkWindow(event->xselectionrequest.requestor),
			  mkSelection(event->xselectionrequest.selection,
				      event->xselectionrequest.target,
				      event->xselectionrequest.property)));
#else
  return mkconstr(rescno,3,
		  mkint(event->xselectionrequest.time),
		  mkWindow(event->xselectionrequest.requestor),
		  mkSelection(event->xselectionrequest.selection,
			      event->xselectionrequest.target,
			      event->xselectionrequest.property));
#endif
}

/*  Selection */
static PTR
putSelectionNotifyEvent(event)
XEvent *event;
{
#ifdef HBC
  return mkconstr1(rescno,mkpair(
		   mkint(event->xselection.time),
		   mkSelection(event->xselection.selection,
			       event->xselection.target,
			       event->xselection.property)));
#else
  return mkconstr(rescno,2,
		   mkint(event->xselection.time),
		   mkSelection(event->xselection.selection,
			       event->xselection.target,
			       event->xselection.property));
#endif
}

static PTR
putStructEvent(event)
XEvent *event;
{
  return mkconstr1(rescno,mkint(event->xmap.window));
}

#if defined (__ANSI_CPP__) || defined(ansi) || defined(__GNUC__)
#define EVENT(eventtype,args) case eventtype: if(debug) fprintf(stderr,#eventtype); rescno = XE##eventtype; rs = args; break
#else
#define EVENT(eventtype,args) case eventtype: if(debug) fprintf(stderr,"eventtype "); rescno = XE/**/eventtype; rs = args; break
#endif

#define Any	    mkconstr0(rescno);
#define Button	    putButtonEvent(event)
#define Motion	    putMotionEvent(event)
#define Key	    putKeyEvent(event)
#define Destroy	    putDestroyWindowEvent(event)
#define Resize	    putResizeRequestEvent(event)
#define Expo	    putExposeEvent(event)
#define GfxExpo	    putGraphicsExposeEvent(event)
#define Crossing    putCrossingEvent(event)
#define FocusChange putFocusChangeEvent(event)
#define Visibility  putVisibilityEvent(event)
#define Configure   putConfigureEvent(event)
#define Client      putClientMessageEvent(event)
#define Struct      putStructEvent(event)

PTR
encodeEvent(event)
     XEvent *event;
{
#if 0
  PTR win, pair;
#else
  PTR rs=0;
#endif
  
  if (debug) fprintf(stderr, "Event %lx ", event->xany.window);
  switch(event->type) {
    EVENT(KeyPress,Key);
    EVENT(KeyRelease,Key);
    EVENT(ButtonPress,Button);
    EVENT(ButtonRelease,Button);
    EVENT(MotionNotify,Motion);
    EVENT(EnterNotify,Crossing);
    EVENT(LeaveNotify,Crossing);
    EVENT(FocusIn,FocusChange);
    EVENT(FocusOut,FocusChange);
    EVENT(KeymapNotify,Any);
    EVENT(Expose,Expo);
    EVENT(GraphicsExpose,GfxExpo);
    EVENT(NoExpose,Any);
    EVENT(VisibilityNotify,Visibility);
    EVENT(CreateNotify,Struct);
    EVENT(DestroyNotify,Struct);
    EVENT(UnmapNotify,Struct);
    EVENT(MapNotify,Struct);
    EVENT(MapRequest,Struct);
    EVENT(ReparentNotify,Any);
    EVENT(ConfigureNotify,Configure);
    EVENT(ConfigureRequest,Any);
    EVENT(GravityNotify,Any);
    EVENT(ResizeRequest,Resize);
    EVENT(CirculateNotify,Any);
    EVENT(CirculateRequest,Any);
    EVENT(PropertyNotify,Any);
    EVENT(SelectionClear,putSelectionClearEvent(event));
    EVENT(SelectionRequest,putSelectionRequestEvent(event));
    EVENT(SelectionNotify,putSelectionNotifyEvent(event));
    EVENT(ColormapNotify,Any);
    EVENT(ClientMessage,Client);
    EVENT(MappingNotify,Any);
  }
  switch(event->type) { /* special hack: don't catch these */
  case ConfigureRequest: {
    XConfigureRequestEvent *e=(XConfigureRequestEvent*)event;
    XWindowChanges wc;
    wc.x=e->x;
    wc.y=e->y;
    wc.width=e->width;
    wc.height=e->height;
    wc.border_width=e->border_width;
    wc.sibling=e->above;
    wc.stack_mode=e->detail;
    XConfigureWindow(e->display,e->window,e->value_mask,&wc);
    break;
  }
  case CirculateRequest: {
    XCirculateRequestEvent *e=(XCirculateRequestEvent*)event;
    if(e->place==PlaceOnTop)
      XRaiseWindow(e->display,e->window);
    else
      XLowerWindow(e->display,e->window);
    break;
  }    
  case ResizeRequest: {
    XResizeRequestEvent *e=(XResizeRequestEvent*)event;
    XResizeWindow(e->display,e->window,e->width,e->height);
    break;
  }
  /* default: */
  }
  if (debug) fprintf(stderr, "\n");
#if 0
  win = mkint(event->xany.window);
  pair = mkpair(win, rs);
  return pair;
#else
  return rs;
#endif
}
